public static class EMCCVSet extends EMCCVMenuElement {
  public int buttonwidth = 70;
  public int probability = 75;
  public boolean visible = true;
  public boolean selected = false;
  public boolean alwayson = false;
  public boolean randomizecolor = true;
  public ArrayList<String> aspects = new ArrayList<String>();

  public EMCCVAspect activeAspect;

  // SCROLLING
  public int position  = 0; // for scrolling
  public int offset    = 0;   // for scrolling
  public int capacity  = 0;  // updated while building aspect menu. How many items fit onto the screen?
  public boolean scrollable = false;
  /*public int scrollbar_h = 0;
  public int handle_h    = 0;
  public int handle_y    = IMG_Y;*/

  // TOOLS
  public boolean colorlink  = false;
  public boolean colorchain = false;
  public boolean collapse  = true;

  @Override
    public void toggleState() {
    selected = !selected;
  }

  @Override
    public boolean isActive() {
    if (logging && logInfo) log.append(getTime() + "[DRAW][IS ACTIVE][SET] id - " + this.id + "\n");

    if (this.disabled) return false;

    if (this.checked == 0) {
      boolean bResult = true;

      if (this.parent != null && !this.parent.equals("") && !this.parent.contains(" " + this.id + " ")) {      
        bResult = checkParents(this.parent);
      }

      bResult = bResult && !this.disabled;

      if (bResult) {
        checked = 1; // visible
      } else {
        checked = -1; // not visible
      }
    }
    return this.checked == 1;
  }

  @Override
    public void randomize() {
    for (String aspectID : this.aspects) {
      EMCCVAspect aspect = (EMCCVAspect) elements.get(aspectID);

      if (aspect != null) {
        aspect.randomize();

        int r = random.nextInt(100);
        if (aspect.alwayson || r <= aspect.probability) {
          aspect.visible = true;
        } else {
          aspect.visible = false;
        }
      }
    }
  }

  public void drawAspectMenu(PApplet pa) {
    int y = ASPECTMENU_Y; // first position

    this.capacity = (pa.height - ASPECTMENU_Y) / ASPECTMENU_D;

    // DRAWING ASPECT MENU
    for (int i = 0; i < aspects.size(); i++) {
      
      // only draw the element if it's not scrolled off the screen
      if (i >= this.offset) {
        String aspectID = aspects.get(i);
        EMCCVAspect aspect = (EMCCVAspect) elements.get(aspectID);

        if (aspect != null && !aspect.incognito && !(aspect.isDisabled() && TOOL_COLLAPSE.active)) {

          // CHECKBOX
          if (aspect.isDisabled()) {
            pa.fill(GUI_DISABLED);
          } else {
            pa.fill(GUI_ENABLED);
          }
          pa.rect(ASPECTMENU_X, y, CONTROLSIZE, CONTROLSIZE, MENURADIUS); 

          // CHECKMARK
          if (aspect.visible) {
            if (aspect.isDisabled() || aspect.alwayson) {
              pa.tint(GUI_SELECTED_DIS);
            } else {
              pa.tint(GUI_SELECTED);
            }
            pa.image(ICON_CHECK, ASPECTMENU_X, y);
          }

          // CHECKBOX into buttonlist
          EMCCVButton checkbox = new EMCCVButton(ASPECTMENU_X, y, CONTROLSIZE, CONTROLSIZE, aspect);
          buttonList.add(checkbox);

          // COLOR CHAINLINK
          if (TOOL_COLORCHAIN.active && aspect.colorizable) {
            if (aspect.colorchain) {
              pa.tint(EMCCVTool.ICON_ACTIVE);
            } else {
              pa.tint(GUI_SELECTED);
            }
            pa.image(ICON_LINK, LINK_X, y);

            // BUTTON (LINK) into buttonlist
            EMCCVButton link = new EMCCVButton(LINK_X, y, CONTROLSIZE, CONTROLSIZE, "link", aspect);
            buttonList.add(link);
          }

          // COG WHEEL
          if (aspect.getNumberOfChildren() > 1) {
            if (aspect.isDisabled()) {
              pa.tint(GUI_SELECTED_DIS);
            } else {
              pa.tint(GUI_SELECTED);
            }
            int tempX = LINK_X; // position depends on whether linklock is activated.
            if (TOOL_COLORCHAIN.active) {
              tempX = COG_X;
            }
            pa.image(ICON_SETTINGS, tempX, y);

            // BUTTON (COG) into buttonlist
            EMCCVButton cog = new EMCCVButton(tempX, y, CONTROLSIZE, CONTROLSIZE, "cog", aspect);
            buttonList.add(cog);
          }

          // TITLE
          if (aspect.isDisabled()) {
            pa.fill(GUI_TEXT_DIS);
          } else {
            pa.fill(GUI_TEXT);
          }
          pa.text(aspect.title, getAspectTitleX(), y-2, getAspectTitleW(), CONTROLSIZE);

          // COLORPICKER
          if (aspect.colorizable && (aspect.colormaster == null || aspect.colormaster.equals(""))) {
            pa.fill(aspect.colour);
            pa.stroke(pa.hue(aspect.colour), pa.saturation(aspect.colour), pa.brightness(aspect.colour)+15);
            pa.rect(CP_X, y, CONTROLSIZE-2, CONTROLSIZE-2, MENURADIUS);

            // COLORPICKER into buttonlist
            EMCCVButton cp = new EMCCVButton(CP_X, y, CONTROLSIZE, CONTROLSIZE, "cp", aspect);
            buttonList.add(cp);
          }
          pa.noStroke();

          y += ASPECTMENU_D;

          // STYLES
          if (aspect == this.activeAspect && aspect.getNumberOfChildren() > 1) {
            for (Map.Entry<String, EMCCVStylegroup> entry : aspect.stylegroups.entrySet()) {

              // separator
              pa.fill(0, 0, 100, 50);
              pa.rect(0, y-(ASPECTMENU_D-CONTROLSIZE)/2, ASPECTMENU_W, 1);
              y++;

              for (Map.Entry<String, EMCCVStyle> style : entry.getValue().styles.entrySet()) {

                // lighter background color
                pa.fill(0, 0, 255, 15);
                pa.rect(0, y-(ASPECTMENU_D-CONTROLSIZE)/2, ASPECTMENU_W, ASPECTMENU_D);

                EMCCVStyle emccvStyle = (EMCCVStyle) style.getValue();
                if (emccvStyle.isDisabled()) {
                  pa.fill(GUI_DISABLED);
                } else {
                  pa.fill(GUI_ENABLED);
                }
                pa.ellipse(getAspectTitleX()-19, y, CONTROLSIZE, CONTROLSIZE);

                if (emccvStyle.selected) {
                  if (emccvStyle.isDisabled()) {
                    pa.fill(GUI_SELECTED_DIS);
                  } else {
                    pa.fill(GUI_SELECTED);
                  }
                  int size = CONTROLSIZE-8;
                  pa.ellipse(getAspectTitleX()-19+((CONTROLSIZE-size)/2), y+((CONTROLSIZE-size)/2), size, size);
                }

                // TITLE
                if (emccvStyle.isDisabled()) {
                  pa.fill(GUI_TEXT_DIS);
                } else {
                  pa.fill(GUI_TEXT);
                }
                if (emccvStyle.title != null) {
                  pa.text(emccvStyle.title, getAspectTitleX(), y-2, getAspectTitleW(), CONTROLSIZE);
                } else {
                  pa.text("unknown", getAspectTitleX(), y-2, getAspectTitleW(), CONTROLSIZE);
                }


                // BUTTON into buttonlist
                EMCCVButton stylebutton = new EMCCVButton(getAspectTitleX()-19, y, CONTROLSIZE, CONTROLSIZE, emccvStyle);
                buttonList.add(stylebutton);

                y += ASPECTMENU_D;
              }
            }
            // separator
            pa.fill(0, 0, 100, 50);
            pa.rect(0, y-(ASPECTMENU_D-CONTROLSIZE)/2, ASPECTMENU_W, 1);
            y++;
          }
        }
      }
    }
    
    

    // SCROLLING SETUP
    if (this.getNumberOfPositions() <= this.capacity) {
      this.offset = 0;
      this.scrollable = false;
    } else {
      this.scrollable = true;
      pa.fill(GUI_ENABLED);
      pa.rect(ASPECTMENU_W-32, IMG_Y, 32, 32*4);
      pa.tint(GUI_SELECTED);
      pa.image(ICON_FIRST, ASPECTMENU_W-28, IMG_Y+4);
      pa.image(ICON_UP,    ASPECTMENU_W-28, IMG_Y+36);
      pa.image(ICON_DOWN,  ASPECTMENU_W-28, IMG_Y+68);
      pa.image(ICON_LAST,  ASPECTMENU_W-28, IMG_Y+100);
    }
  }
  
  public int getNumberOfPositions() {
    int result = 0;
    
    // count available aspects (on and offscreen)
    for (String aspectid : this.aspects) {
      if (elements.get(aspectid) != null && (!this.collapse || !elements.get(aspectid).isDisabled())) {
        result++;
      }
    }
    
    if (this.activeAspect != null) {
      result = result + this.activeAspect.getNumberOfChildren();
    }
    return result;
  }
}