import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import java.text.SimpleDateFormat; 
import java.util.Date; 
import java.util.Random; 
import java.util.Map.Entry; 
import java.util.Map; 
import java.util.LinkedHashMap; 
import java.util.TreeMap; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class EMCCV_Prototype extends PApplet {









String EMCCV_VERSION = "v4.2";
public static EMCCVModule activeModule = new EMCCVModule();
String CONFIGFILE;
String DIR_ENTITIES;
String DIR_MODULES = "modules/";
String DIR_EXT;
String DIR_MOD;
String DIR_IMAGES;
String DIR_DATA;

int mode = 0; // dashboard
final int DASH    = 0;
final int LOAD    = 1;
final int SAVE    = 2;
final int CAPTURE = 3;
final int RELOAD  = 4;
final int CP      = 5;
final int PLAY    = 6;

PGraphics pg;
PGraphics pgExport;

static Random random = new Random();

static XML xml;
static ArrayList<String> moduleids = new ArrayList();
static HashMap<String, EMCCVMenuElement> elements = new HashMap<String, EMCCVMenuElement>();
static Map<String, EMCCVMenuElement> randomKeys = new TreeMap<String, EMCCVMenuElement>();
static ArrayList<EMCCVFile> fileLayers = new ArrayList<EMCCVFile>();
static HashMap<String, PImage> graphics = new HashMap<String, PImage>();
static ArrayList COLOURS_50BRI = new ArrayList();

PFont OpenSans12;
PFont SansSerif12;

static boolean logging    = false;
static boolean logInitXML = logging && true;
static boolean logInfo    = logging && true;
static boolean logGui     = logging && true;
static boolean logActions = logging && false;
static boolean logSuccess = logging && false;
public static StringBuilder log = new StringBuilder();
static String logDirectory = "";
static String logline = "------------------------------------\n";
static String indent  = "||||||||||||||||||| ";

public void setup() {
  
  surface.setResizable(true);
  setLogDirectory();
  background(GUI_ACTIVE);
  
  //imageMode(CENTER);
  //textAlign(CENTER);
  //fill(GUI_TEXT);
  //text("loading...", width/2-100, height/2+50, 200, 50);
  //IMG_LOGO = loadImage(dataPath("images/gui/logo.png"));
  //image(IMG_LOGO, width/2, height/2);
  //imageMode(CORNER);
  //textAlign(CENTER, LEFT);
  
  colorMode(HSB, 255);
  ellipseMode(CORNER);
  noLoop();
  
  thread("loadIcons");
  thread("loadTools");
  thread("loadPalette");

  OpenSans12  = loadFont("OpenSans-12.vlw");
  SansSerif12 = loadFont("SansSerif-12.vlw");

  if (logging && logInfo) log.append(getTime() + "====================================\n");
  if (logging && logInfo) log.append(getTime() + " SETUP COMPLETE!\n");
  if (logging && logInfo) log.append(getTime() + "====================================\n");
  
  // Just in case we crash before the first draw finishes:
  log = printLog(log, "SETUP/");
}

public void draw() {
  cursor(WAIT);
  noStroke();

  if (redrawSetMenu) drawMenuSets();
  if (redrawToolbar) drawToolbar();
  
  switch (mode) {
    case DASH:
      redrawAspectMenu = false;
      redrawModel      = false;
      drawModuleMenu();
      break;
      
    case CAPTURE:
      mode = PLAY;
      drawAlert("Image saved in the modules' images folder");
      TOOL_CAPTURE.active = false;
      redrawAspectMenu    = false;
      redrawToolbar       = true;
      frameRate(0.3f);
      loop();
      redraw();
      break;
    case SAVE:
      mode = PLAY;
      drawAlert("Entity saved in the modules' entities folder");
      TOOL_SAVE.active = false;
      redrawAspectMenu = false;
      redrawToolbar    = true;
      frameRate(0.5f);
      loop();
      redraw();
      break;
    case RELOAD:
      mode = PLAY;
      drawAlert("Reloading...");
      TOOL_RELOAD.active = false;
      redrawSetMenu      = true;
      redrawAspectMenu   = false;
      redrawToolbar      = true;
      frameRate(0.5f);
      loop();
      redraw();
      break;
    case CP:
      drawColorPicker();
      break;
    case LOAD:
      redrawAspectMenu = false;
      redrawModel = false;
      loadThumbnails();
      break;
    default:
      if (refreshModel) {
        redrawAspectMenu = true;
        redrawModel      = true;
        refreshModel     = false;
        loop();
      } else {
        noLoop();
        frameRate(60);
        redrawAspectMenu = true;
        redrawModel = true;
      }
  }
  
  try {
    if (redrawAspectMenu && activeModule != null) drawAspectMenu();
    if (redrawModel) {
      updateView();
      fill(GUI_VIEW);
      rect(IMG_X, IMG_Y, width-ASPECTMENU_W, view_H);
      preparePG();
      resetCheckedMarker();
      //tint(0, 0, 0, 50);
      //image(pg, IMG_X+2*MARGIN, IMG_Y-MARGIN, view_W, view_H);
      noTint();
      image(pg, IMG_X, IMG_Y, view_W, view_H);
    }
    redrawModel = false;
    redrawSetMenu = false;
    redrawAspectMenu = false;
  } catch (Exception e) {
    log.append(getTime() + " ============================================================ \n");
    log.append(indent    + " ========= !!! ERROR DURING DRAWING PROCESS !!! ============= \n");
    log.append(indent    + " ========= !!!          Error #0001         !!! ============= \n");
    log.append(indent    + " ========= !!!       Reloading module       !!! ============= \n");
    log.append(indent    + " ============================================================ \n");
    redraw();
  }

  log = printLog(log, "DRAW/");
}

public void preparePG() {
  updateView();
  pg = createGraphics(activeModule.w, activeModule.h);
  pg.beginDraw();

  pg.colorMode(HSB, 255);
  pg.background(255, 255, 255, 0);
  pg.clear();
  pg.noStroke();
  pg.noTint();

  for (EMCCVFile file : fileLayers) {
    if (logging && logInfo) {
      log.append(getTime() + "[DRAW][LAYER]: ").append(file.layer).append(" - ").append(file.src).append("\n");  
      log.append(getTime() + "                 xmlParent: ").append(file.xmlParent).append("\n");
    }

    if (file.isActive()) {
      pg.blendMode(BLEND);
      
      // add image file to graphics HashMap if not there already
      if (graphics.get(sketchPath(DIR_DATA) + file.src) == null) {
        try {
          PImage img = loadImage(sketchPath(DIR_DATA) + file.src);
          graphics.put(sketchPath(DIR_DATA) + file.src, img);
          if (logging && logSuccess) log.append(getTime() + "[DRAW][FILE ADDED]: ").append(file.src).append("\n");
        } 
        catch (Exception e) {
          log.append(getTime() + "[DRAW] ==> FILE NOT FOUND: ").append(file.src).append("\n");
        }
      }

      // CHECKING FOR COLORMASTER
      if (file.colormaster != null && !file.colormaster.equals("")) {
        try {
          pg.tint(file.getColormaster().getColour());
        } 
        catch (Exception e) {
          log.append(getTime() + "[ERROR] Can't find colormaster." + e);
        }
      } else if (file.colorizable) {
        pg.tint(file.getColour());
      } else {
        pg.noTint();
      }
      if (file.blendmode != null && file.blendmode.equals("multiply")) {
        pg.blendMode(MULTIPLY);
      }
      try { //<>//
        // draw image
        if (file.w != 0 && file.h != 0) {
          // flip
          if (file.flipped) {
            pg.pushMatrix();
            pg.translate(graphics.get(sketchPath(DIR_DATA) + file.src).width,0);
            pg.scale(-1,1);
          }
          pg.image(graphics.get(sketchPath(DIR_DATA) + file.src), file.x, file.y, file.w, file.h);
          // unflip
          if (file.flipped) {
            pg.popMatrix();
          }
        } else {
          // flip
          if (file.flipped) {
            pg.pushMatrix();
            pg.translate(graphics.get(sketchPath(DIR_DATA) + file.src).width,0);
            pg.scale(-1,1);
          }
          pg.image(graphics.get(sketchPath(DIR_DATA) + file.src), file.x, file.y);
          // unflip
          if (file.flipped) {
            pg.popMatrix();
          }
        }
      } 
      catch (Exception e) {
        log.append("\n" + getTime() + "[ERROR] FILE COULD NOT BE FOUND: ").append(file.src).append("\n");
        println("\n" + getTime() + "[ERROR] FILE COULD NOT BE FOUND: " + file.src + "\n");
      }
    }
  }
  pg.endDraw();
  pgExport = createGraphics(activeModule.w, activeModule.h+20);
  pgExport.beginDraw();
  pgExport.image(pg, 0, 0);
  pgExport.noTint();
  pgExport.noStroke();

  int footerColor = pg.get(20, activeModule.h-20);
  pgExport.fill(footerColor);
  pgExport.rect(-1, activeModule.h-1, activeModule.w+2, 21);
  pgExport.fill(0, 0, 0, 75);
  pgExport.rect(-1, activeModule.h-1, activeModule.w+2, 22);
  pgExport.fill(255, 95);
  String credit = "Model: " + activeModule.title + " v" + activeModule.version;
  pgExport.textAlign(RIGHT, CENTER);
  pgExport.textFont(SansSerif12, 12);
  pgExport.text(credit, 0, activeModule.h, activeModule.w-5, 20);
  credit = "EMCCV Prototype " + EMCCV_VERSION;
  pgExport.textAlign(LEFT, CENTER);
  pgExport.text(credit, 5, activeModule.h, activeModule.w-5, 20);
  pgExport.noFill();

  pgExport.blendMode(BLEND);
  pgExport.image(IMG_CREDITS, 15, activeModule.h-50, 35, 35);
  pgExport.endDraw();
}

public void resetCheckedMarker() {
  for (Map.Entry<String, EMCCVMenuElement> element : elements.entrySet()) {
    EMCCVMenuElement e = (EMCCVMenuElement) element.getValue();    
    e.checked = 0;
  }
}

public void initXML(String moduleid) {
  
  randomKeys.clear();
  setList.clear();
  elements.clear();
  fileLayers.clear();
  graphics.clear();

  if (logging && logInitXML) {
    log.append(getTime() + "====================================\n");
    log.append(getTime() + " INITIALIZING XML\n");
    log.append(getTime() + "====================================\n");
  }
  
  activeModule    = new EMCCVModule();
  activeModule.id = moduleid;

  CONFIGFILE   = "modules/" + activeModule.id + "/config.xml";
  DIR_EXT      = "modules/" + activeModule.id + "/ext/";
  DIR_MOD      = "modules/" + activeModule.id + "/mod/";
  DIR_ENTITIES = "modules/" + activeModule.id + "/entities/";
  DIR_IMAGES   = "modules/" + activeModule.id + "/images/";
  DIR_DATA     = "modules/" + activeModule.id + "/data/";

  setLogDirectory();
  
  try {
    xml = loadXML(CONFIGFILE);
    activeModule.title    = xml.getString("title");
    activeModule.version  = xml.getString("version");
    activeModule.sDesc    = xml.getString("sDesc");
    activeModule.authorid = xml.getString("authorid");
    activeModule.w        = xml.getInt("width");
    activeModule.h        = xml.getInt("height");
    activeModule.logo     = loadImage(sketchPath(DIR_DATA + "logo.png"));
    activeModule.randomizecolor = xml.getString("randomizecolor") == null || xml.getString("randomizecolor").equals("true");
  
    if (logging && logInitXML) {
      log.append(getTime() + " WIDTH: ").append(activeModule.w).append(", HEIGHT: ").append(activeModule.h).append("\n");
      log.append(getTime() + logline);
    }
    
    EMCCVSet infoset = new EMCCVSet();
    infoset.id = "setInfo";
    infoset.title = "INFO";
    infoset.buttonwidth = 60;
    setList.add(infoset);
    if (logging && logInitXML) log.append(getTime() + " INFO tab added.\n");
  
    XML[] xmlSets = xml.getChildren("set");
  
    for (XML xmlSet : xmlSets) {
      EMCCVSet set = new EMCCVSet();
  
      // SET attributes
      set.id          = processStringAttribute(xmlSet, "id");
      set.title       = processStringAttribute(xmlSet, "title", "unknown title");
      set.disabled    = processBooleanAttribute(xmlSet, "disabled");
      set.parent      = processStringAttribute(xmlSet, "parent");
      set.colormaster = processStringAttribute(xmlSet, "colormaster");
      set.colour      = processColorAttribute(xmlSet, "color", color(255, 255, 255));
      set.colorizable = processBooleanAttribute(xmlSet, "colorizable");
      set.randomizecolor = processBooleanAttribute(xmlSet, "randomizecolor", activeModule.randomizecolor);
      set.x           = processIntAttribute(xmlSet, "x");
      set.y           = processIntAttribute(xmlSet, "y");
      set.w           = processIntAttribute(xmlSet, "w");
      set.h           = processIntAttribute(xmlSet, "h");
      set.buttonwidth = processIntAttribute(xmlSet, "buttonwidth");
      set.layer       = processFloatAttribute(xmlSet, "layer");
      set.alwayson    = processBooleanAttribute(xmlSet, "alwayson");
      set.randomize   = processStringAttribute(xmlSet, "randomize");
      set.probability = processIntAttribute(xmlSet, "probability", 75);
      set.blendmode   = processStringAttribute(xmlSet, "blendmode");
      set.flipped     = processBooleanAttribute(xmlSet, "flipped");
      set.visible     = processBooleanAttribute(xmlSet, "visible", false);
      set.incognito   = processBooleanAttribute(xmlSet, "incognito", false);
  
      if (logging && logInitXML) log.append(getTime() + logline);
  
      // SET > Aspects
      XML[] xmlAspects = xmlSet.getChildren("aspect");
      for (XML xmlAspect : xmlAspects) {
        EMCCVAspect aspect = new EMCCVAspect();
  
        // ASPECT attributes
        aspect.id          = processStringAttribute(xmlAspect, "id");      
        aspect.title       = processStringAttribute(xmlAspect, "title", "unknown title");
        aspect.disabled    = processBooleanAttribute(xmlAspect, "disabled");
        aspect.parent      = processStringAttribute(xmlAspect, "parent");
        aspect.colormaster = processStringAttribute(xmlAspect, "colormaster", set.colormaster);
        aspect.colour      = processColorAttribute(xmlAspect, "color", set.colour);
        aspect.colorizable = processBooleanAttribute(xmlAspect, "colorizable", set.colorizable);
        aspect.randomizecolor = processBooleanAttribute(xmlAspect, "randomizecolor", set.randomizecolor);
        aspect.x           = processIntAttribute(xmlAspect, "x", set.x);
        aspect.y           = processIntAttribute(xmlAspect, "y", set.y);
        aspect.w           = processIntAttribute(xmlAspect, "w", set.w);
        aspect.h           = processIntAttribute(xmlAspect, "h", set.h);
        aspect.visible     = processBooleanAttribute(xmlAspect, "visible", set.visible || processBooleanAttribute(xmlAspect, "alwayson", set.alwayson));
        aspect.incognito   = processBooleanAttribute(xmlAspect, "incognito", false);
        aspect.layer       = processFloatAttribute(xmlAspect, "layer", set.layer);
        aspect.alwayson    = processBooleanAttribute(xmlAspect, "alwayson", set.alwayson);
        aspect.randomize   = processStringAttribute(xmlAspect, "randomize");
        aspect.probability = processIntAttribute(xmlAspect, "probability", set.probability);
        aspect.blendmode   = processStringAttribute(xmlAspect, "blendmode", set.blendmode);
        aspect.flipped     = processBooleanAttribute(xmlAspect, "flipped", set.flipped);
        aspect.xmlParent   = set.id;
  
        if (aspect.alwayson) {
          aspect.visible = true;
        }
  
        if (logging && logInitXML) log.append(getTime() + logline);
  
        // SET > ASPECT > Styles
        XML[] xmlStyles = xmlAspect.getChildren("style");
        for (XML xmlStyle : xmlStyles) {
          EMCCVStyle style = new EMCCVStyle();
  
          // STYLE attributes
          style.id          = processStringAttribute(xmlStyle, "id");
          style.title       = processStringAttribute(xmlStyle, "title", "unknown title");
          style.disabled    = processBooleanAttribute(xmlStyle, "disabled");
          style.parent      = processStringAttribute(xmlStyle, "parent");
          style.colormaster = processStringAttribute(xmlStyle, "colormaster", aspect.colormaster);
          style.colour      = processColorAttribute(xmlStyle, "color", aspect.colour);
          style.colorizable = processBooleanAttribute(xmlStyle, "colorizable", aspect.colorizable);
          style.x           = processIntAttribute(xmlStyle, "x", aspect.x);
          style.y           = processIntAttribute(xmlStyle, "y", aspect.y);
          style.w           = processIntAttribute(xmlStyle, "w", aspect.w);
          style.h           = processIntAttribute(xmlStyle, "h", aspect.h);
          style.incognito   = processBooleanAttribute(xmlStyle, "incognito", false);
          style.layer       = processFloatAttribute(xmlStyle, "layer", aspect.layer);
          style.stylegroup  = processStringAttribute(xmlStyle, "stylegroup", "default");
          style.blendmode   = processStringAttribute(xmlStyle, "blendmode", aspect.blendmode);
          style.flipped     = processBooleanAttribute(xmlStyle, "flipped", aspect.flipped);
          style.selected    = processBooleanAttribute(xmlStyle, "selected", false);
          style.xmlParent   = aspect.id;
  
          // if the stylegroup doesn't exist -> create it
          if (aspect.stylegroups.get(style.stylegroup) == null) {
            EMCCVStylegroup stylegroup = new EMCCVStylegroup(style.stylegroup);
            aspect.stylegroups.put(stylegroup.title, stylegroup);
          }
  
          if (logging && logInitXML) log.append(getTime() + logline);
  
          // SET > ASPECT > STYLE > File
          XML[] xmlFiles = xmlStyle.getChildren("file");
          for (XML EMCCVFile : xmlFiles) {
            EMCCVFile file = new EMCCVFile();
  
            // FILE attributes
            file.parent      = processStringAttribute(EMCCVFile, "parent");
            file.colormaster = processStringAttribute(EMCCVFile, "colormaster", style.colormaster);
            file.colour      = processColorAttribute(EMCCVFile, "color", style.colour);
            file.colorizable = processBooleanAttribute(EMCCVFile, "colorizable", style.colorizable);
            file.x           = processIntAttribute(EMCCVFile, "x", style.x);
            file.y           = processIntAttribute(EMCCVFile, "y", style.y);
            file.w           = processIntAttribute(EMCCVFile, "w", style.w);
            file.h           = processIntAttribute(EMCCVFile, "h", style.h);
            file.layer       = processFloatAttribute(EMCCVFile, "layer", style.layer);
            file.src         = processStringAttribute(EMCCVFile, "src");
            file.blendmode   = processStringAttribute(EMCCVFile, "blendmode", style.blendmode);
            file.flipped     = processBooleanAttribute(EMCCVFile, "flipped", style.flipped);
            file.xmlParent   = style.id;
  
            // add file into ArrayList
            style.files.add(file);
            // add file into fileLayers ArrayList
            if (fileLayers.size() == 0 || file.layer >= fileLayers.get(fileLayers.size() - 1).layer) {
              fileLayers.add(file);
            } else {
              for (int i = 0; i < fileLayers.size(); i++) {
                if (fileLayers.get(i).layer > file.layer) {
                  fileLayers.add(i, file);
                  break;
                }
              }
            }
  
            if (logging && logInitXML) log.append(getTime() + logline);
          } // FILES DONE!
          if (logging && logInitXML) log.append(getTime() + " [INITXML] FILES added.\n");
          
          // add style into HashMaps and set the first one to selected
          if (aspect.stylegroups.get(style.stylegroup).styles.size() == 0) {
            style.selected = true;
          } else if (style.selected) {
            aspect.stylegroups.get(style.stylegroup).resetStyles();
          }
          aspect.stylegroups.get(style.stylegroup).styles.put(style.id, style);
          if (elements.get(style.id) != null) log.append(getTime() + " [INITXML] STYLE with ID = " + style.id + " already exists and will be overwritten!\n");
          elements.put(style.id, style);
        }
        if (logging && logInitXML) log.append(getTime() + " [INITXML] STYLES added.\n");
  
        // add aspect into HashMaps
        set.aspects.add(aspect.id);
        if (elements.get(aspect.id) != null) log.append(getTime() + " [INITXML] ASPECT with ID = " + aspect.id + " already exists and will be overwritten!\n");
        elements.put(aspect.id, aspect);
        if (aspect.randomize != null && !aspect.randomize.isEmpty()) {
          if (randomKeys.get(aspect.randomize) != null) log.append(getTime() + " [INITXML] ASPECT.randomize = " + aspect.randomize + " already exists and will be overwritten!\n");
          randomKeys.put(aspect.randomize, aspect);
        }
      }
      if (logging && logInitXML) log.append(getTime() + " [INITXML] ASPECTS added.\n");
      
      // add set into HashMap
      if (elements.get(set.id) != null) log.append(getTime() + " [INITXML] SET with ID = " + set.id + " already exists and will be overwritten!\n");
      elements.put(set.id, set);
      if (logging && logInitXML) log.append(getTime() + " [INITXML] SET " + set.id + " added to elements list.\n");
      
      if (set.randomize != null && !set.randomize.isEmpty()) {
        if (randomKeys.get(set.randomize) != null) log.append(getTime() + " [INITXML] SET.randomize = " + set.randomize + " already exists and will be overwritten!\n");
        randomKeys.put(set.randomize, set);
      }
      setList.add(set);
      if (logging && logInitXML) log.append(getTime() + " [INITXML] SET " + set.id + " added to setList.\n");
      
      activeModule.sets.add(set.id);
      if (logging && logInitXML) log.append(getTime() + " [INITXML] SET " + set.id + " added to activeModule.sets.\n");
    }
    if (logging && logInitXML) log.append(getTime() + " [INITXML] SETS added: " + setList.size()+ "\n");
    setList.get(0).selected = true;
    activeModule.activeSet = setList.get(0);
    if (logging && logInitXML) log.append(getTime() + " [INITXML] ACTIVE MODULE = " + setList.get(0).id + "\n");
  
    if (logging && logInitXML) {
      log.append(getTime() + "====================================\n");
      log.append(getTime() + " DATA ANALYSIS\n");
      log.append(getTime() + "====================================\n");
      log.append(getTime() + " Number of XML elements: ").append(elements.size()).append("\n");
      log.append(getTime() + " Number of fileLayers: ").append(fileLayers.size()).append("\n");
      log.append(getTime() + "\n");
    }
    log = printLog(log, "XML/");
  
    if (logging && logInitXML) {
      log.append(getTime() + "====================================\n");
      log.append(getTime() + " LIST OF LAYERS\n");
      log.append(getTime() + "====================================\n");
      log.append(getTime() + " Number of fileLayers: ").append(fileLayers.size()).append("\n");
      log.append(getTime() + "\n");
  
      for (EMCCVFile layer : fileLayers) {
        log.append(getTime() + "[LAYER " + layer.layer + "] " + layer.src + "\n");
      }
  
      log = printLog(log, "layers/");
    }
    
    updateView();
  
    int window_W = view_W + ASPECTMENU_W;
    int window_H = displayHeight - 120;
    surface.setSize(window_W, window_H);
  
  } catch (Exception e) {
    log.append(getTime() + "ERROR READING XML FILE\n");
  }
}

private String processStringAttribute(XML xml, String attribute) {
  String result = null;
  if (xml.getString(attribute) != null) {
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(xml.getString(attribute)).append("\n");
    result = xml.getString(attribute).trim().replaceAll(" +", " ");
  } else {
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" not declared.\n");
  }
  return result;
}

private String processStringAttribute(XML xml, String attribute, String inherited) {
  String result = null;
  if (xml.getString(attribute) != null) {
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(xml.getString(attribute)).append("\n");
    result = xml.getString(attribute).trim().replaceAll(" +", " ");
  } else if (inherited != null) {
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(inherited).append(" (inherited)\n");
    result = inherited;
  } else {
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" not declared.\n");
  }
  return result;
}

private boolean processBooleanAttribute(XML xml, String attribute) {
  boolean result = false;
  if (attribute == "colorizable") {
    result = true;
  }
  if (xml.getString(attribute) != null) {
    result = xml.getString(attribute).equals("true");
  }
  if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(result).append("\n");
  return result;
}

private boolean processBooleanAttribute(XML xml, String attribute, boolean inherited) {
  boolean result = false;
  if (attribute == "colorizable") {
    result = true;
  }
  if (xml.getString(attribute) != null) {
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(xml.getString(attribute)).append("\n");
    result = xml.getString(attribute).equals("true");
  } else {
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(inherited).append(" (inherited)\n");
    result = inherited;
  }
  return result;
}

private int processColorAttribute(XML xml, String attribute, int inherited) {
  int result = color(255);
  if (xml.getString(attribute) != null) {
    try {
      result = unhex("FF" + xml.getString(attribute).substring(1));
      if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(hex(result)).append("\n");
    } 
    catch (Exception e) {
      log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(hex(result)).append(" (declaration illegal)\n");
    }
  } else {
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(hex(inherited)).append(" (inherited)\n");
    result = inherited;
  }
  return result;
}

private float processFloatAttribute(XML xml, String attribute) {
  float result = 0;
  if (xml.getString(attribute) != null) {
    result = xml.getFloat(attribute);
  }
  if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(result).append("\n");
  return result;
}

private float processFloatAttribute(XML xml, String attribute, float inherited) {
  float result = 0;
  if (xml.getString(attribute) != null) {
    result = xml.getFloat(attribute);
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(xml.getString(attribute)).append("\n");
  } else {
    result = inherited;
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(inherited).append(" (inherited)\n");
  }
  return result;
}

private int processIntAttribute(XML xml, String attribute) {
  int result = 0;
  if (xml.getString(attribute) != null) {
    result = xml.getInt(attribute);
  }
  if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(result).append("\n");
  return result;
}

private int processIntAttribute(XML xml, String attribute, int inherited) {
  int result = 0;
  if (xml.getString(attribute) != null) {
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(xml.getString(attribute)).append("\n");
    result = xml.getInt(attribute);
  } else {
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(inherited).append(" (inherited)\n");
    result = inherited;
  }
  return result;
}

/**
 * parses the parent attribute and checks to see if all conditions are met
 *
 */
public static boolean checkParents(String parents) {
  boolean bResult = true;

  if (parents != null) {
    // if the parent contains only one word, deal with it.
    if (!(parents.contains(",") || parents.contains(" "))) {
      EMCCVMenuElement parent;
      boolean invert = false;
      
      if (parents.contains("!")) {
        parent = elements.get(parents.substring(1));
        invert = true;
      } else {
        parent = elements.get(parents);
      }
      if (parent != null) {
        bResult = bResult && parent.isActive();
        if (invert) {
          bResult = !bResult;
        }
        if (logging && logInfo) log.append(getTime() + "[DRAW][IS ACTIVE][CHECK SINGLE PARENTS] " + parent.id + " active: " + bResult + "\n");
      } else {
        log.append(getTime() + parents).append(" id not a valid ID.\n");
      }
    } else {
      String[] prerequisites = split(parents, ", ");
      if (logging && logInfo) log.append(getTime() + "[DRAW][IS ACTIVE][CHECK MULTIPLE PARENTS] " + parents + ": " + prerequisites.length + " statement(s)\n");

      // iterate through prerequites
      for (String prereq : prerequisites) {

        // if this prerequisite contains multiple IDs (separated by spaces)...
        if (prereq.contains(" ")) {
          String[] words = split(prereq, " ");

          EMCCVMenuElement tempParent = null;
          boolean tempActive = true;
          String tempOperator = "";
          boolean prevWordWasId = false;

          if (logging && logInfo) log.append(getTime() + "      ");

          for (String word : words) {
            if (word.equals("and") || word.equals("or")) {
              if (prevWordWasId) {
                tempOperator = word;
                prevWordWasId = false;
                if (logging && logInfo) log.append(getTime() + " " + word + " ");
              } else {
                log.append(getTime() + "Expecting ID before operator. Prerequisite ignored.\n");
                break;
              }
            } else {
              // next word should be an ID -> deal with it.
              EMCCVMenuElement parent;
              boolean invert = false;
              
              if (word.contains("!")) {
                parent = elements.get(word.substring(1));
                invert = true;
              } else {
                parent = elements.get(word);
              }
              if (parent != null) {
                // we have a parent object
                if (logInfo && invert) log.append(getTime() + word + " (" + parent.isActive() + ")\n");
                if (logInfo && !invert) log.append(getTime() + word + " (" + !parent.isActive() + ")\n");

                if (tempParent == null) {
                  // this is the first ID in this statement -> memorize
                  tempParent = parent;
                  if (invert) {
                    tempActive = tempActive && !parent.isActive();
                  } else {
                    tempActive = tempActive && parent.isActive();
                  }                  
                  prevWordWasId = true;
                } else {
                  // we have already had an object in this statement
                  // -> we should also have an operator
                  if (!tempOperator.isEmpty()) {
                    // operator OK -> proceed
                    if (tempOperator.equals("and")) {
                      tempActive = tempActive && parent.isActive();
                    } else if (tempOperator.equals("or")) {
                      tempActive = tempActive || parent.isActive();
                    } else {
                      log.append(getTime() + "Expecting operator 'and/or' between IDs. Got ")
                        .append(word)
                        .append(" instead. Prerequisite ignored.\n");
                    }
                    tempOperator = "";
                    prevWordWasId = true;
                  } else {
                    log.append(getTime() + "Got multiple IDs (at ").append(word).append(") but no operator. Prerequisite ignored.\n");
                    break;
                  }
                }
              } else {
                log.append(getTime() + word).append(" id not a valid ID. Prerequisite ignored.\n");
                break;
              }
            }
          }
          bResult = bResult && tempActive;
        }
        if (logging && logInfo) log.append(getTime() + "[DRAW][IS ACTIVE][PREREQ] " + prereq + " (active: " + bResult + ")\n");
      }
    }
  }

  return bResult;
}

public void drawAlert(String message) {
  textAlign(LEFT, TOP);
  fill(GUI_ACTIVE);
  rect(0, IMG_Y, ASPECTMENU_W, height-IMG_Y);
  fill(GUI_TEXT);
  text(message, MARGIN, IMG_Y+MARGIN, ASPECTMENU_W-2*MARGIN, height-IMG_Y-2*MARGIN);
}

public static String getTime() {
  SimpleDateFormat sdf = new SimpleDateFormat("[yyyyMMdd-HH:mm:ss] ");
  return sdf.format(new Date());
}

public static String getDateTime() {
  SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMddHHmmss");
  return sdf.format(new Date());
}

public void loadModuleIDs() {
  moduleids.clear();
  
  File dir = new File(sketchPath(DIR_MODULES));
  File[] files = dir.listFiles();
  if (logging && logInfo) log.append(getTime() + "[LOAD MODULES] Number of files/directories found: " + files.length + "\n");

  moduleids = new ArrayList();

  if (files != null) {
    for (File file : files) {
      if (file.isDirectory()) {
        String filename = file.getName();
        if (logging && logInfo) log.append(getTime() + "[LOAD MODULES] Folder found: " + filename + "\n");
        File config = new File(sketchPath(DIR_MODULES + filename + "/config.xml"));
        if (config.isFile()) {
          moduleids.add(filename);
          if (logging && logInfo) log.append(getTime() + "[LOAD MODULES] Config found.\n");
        }
      }
    }
  }
}

public void loadPalette() {
  COLOURS_50BRI.add(unhex("FFff0000"));
  COLOURS_50BRI.add(unhex("FFff4000"));
  COLOURS_50BRI.add(unhex("FFff8000"));
  COLOURS_50BRI.add(unhex("FFffbf00"));
  COLOURS_50BRI.add(unhex("FFffff00"));
  COLOURS_50BRI.add(unhex("FFbfff00"));
  COLOURS_50BRI.add(unhex("FF80ff00"));
  COLOURS_50BRI.add(unhex("FF40ff00"));
  COLOURS_50BRI.add(unhex("FF00ff00"));
  COLOURS_50BRI.add(unhex("FF00ff40"));
  COLOURS_50BRI.add(unhex("FF00ff80"));
  COLOURS_50BRI.add(unhex("FF00ffbf"));
  COLOURS_50BRI.add(unhex("FF00ffff"));
  COLOURS_50BRI.add(unhex("FF00bfff"));
  COLOURS_50BRI.add(unhex("FF0080ff"));
  COLOURS_50BRI.add(unhex("FF0040ff"));
  COLOURS_50BRI.add(unhex("FF0000ff"));
  COLOURS_50BRI.add(unhex("FF4000ff"));
  COLOURS_50BRI.add(unhex("FF8000ff"));
  COLOURS_50BRI.add(unhex("FFbf00ff"));
  COLOURS_50BRI.add(unhex("FFff00ff"));
  COLOURS_50BRI.add(unhex("FFff00bf"));
  COLOURS_50BRI.add(unhex("FFff0080"));
  COLOURS_50BRI.add(unhex("FFff0040"));
  COLOURS_50BRI.add(unhex("FFffffff"));
  COLOURS_50BRI.add(unhex("FFcccccc"));
  COLOURS_50BRI.add(unhex("FF999999"));
  COLOURS_50BRI.add(unhex("FF666666"));
  COLOURS_50BRI.add(unhex("FF333333"));
  COLOURS_50BRI.add(unhex("FF000000"));
  if (logging && logInfo) log.append(getTime() + "[SETUP] Palettes loaded.\n");
}
public void loadIcons() {
  IMG_LOGO = loadImage(dataPath("images/gui/logo.png"));
  IMG_CREDITS =  loadImage(dataPath("images/gui/credits.png"));
  IMG_GRADIENT_HUE =  loadImage(dataPath("images/gui/gradient_hue.png"));
  IMG_GRADIENT_SAT =  loadImage(dataPath("images/gui/gradient_sat.png"));
  IMG_GRADIENT_BRI =  loadImage(dataPath("images/gui/gradient_bri.png"));
  ICON_CHECK    = loadImage(dataPath("images/icons/16/check.png"));
  ICON_LINK     = loadImage(dataPath("images/icons/16/link.png"));
  ICON_SETTINGS = loadImage(dataPath("images/icons/16/settings.png"));
  ICON_FIRST = loadImage(dataPath("images/icons/24/first.png"));
  ICON_UP    = loadImage(dataPath("images/icons/24/up.png"));
  ICON_DOWN  = loadImage(dataPath("images/icons/24/down.png"));
  ICON_LAST  = loadImage(dataPath("images/icons/24/last.png"));
  if (logging && logInfo) log.append(getTime() + "[SETUP] Icons loaded.\n");
}
public void loadTools() {
  TOOL_LOAD           = new EMCCVTool(MARGIN + (EMCCVTool.ICON_SIZE + PADDING)*0, "Load Entity", loadImage(dataPath("images/icons/32/load.png")));
  TOOL_SAVE           = new EMCCVTool(MARGIN + (EMCCVTool.ICON_SIZE + PADDING)*1, "Save Entity", loadImage(dataPath("images/icons/32/save.png")));
  TOOL_CAPTURE        = new EMCCVTool(MARGIN + (EMCCVTool.ICON_SIZE + PADDING)*2, "Capture Image", loadImage(dataPath("images/icons/32/capture.png")));
  TOOL_COLORLINK      = new EMCCVTool(MARGIN*2 + (EMCCVTool.ICON_SIZE + PADDING)*3, "Color Link", loadImage(dataPath("images/icons/32/colorlink.png")));
  TOOL_COLORCHAIN     = new EMCCVTool(MARGIN*2 + (EMCCVTool.ICON_SIZE + PADDING)*4, "Color Chain", loadImage(dataPath("images/icons/32/colorchain.png")));
  TOOL_COLLAPSE       = new EMCCVTool(MARGIN*2 + (EMCCVTool.ICON_SIZE + PADDING)*5, "Collapse", loadImage(dataPath("images/icons/32/collapse.png")));
  TOOL_RANDOMIZE      = new EMCCVTool(MARGIN*2 + (EMCCVTool.ICON_SIZE + PADDING)*6, "Randomize", loadImage(dataPath("images/icons/32/randomize.png")));
  TOOL_RANDOMIZECOLOR = new EMCCVTool(MARGIN*2 + (EMCCVTool.ICON_SIZE + PADDING)*7, "Randomize Color", loadImage(dataPath("images/icons/32/randomizecolor.png")));
  TOOL_COLORSHIFT     = new EMCCVTool(MARGIN*2 + (EMCCVTool.ICON_SIZE + PADDING)*8, "Color Shoft", loadImage(dataPath("images/icons/32/colorshift.png")));
  TOOL_GLOBALMODE     = new EMCCVTool(MARGIN*2 + (EMCCVTool.ICON_SIZE + PADDING)*9, "Global Mode", loadImage(dataPath("images/icons/32/globalmode.png")));
  TOOL_LOG            = new EMCCVTool(MARGIN*3 + (EMCCVTool.ICON_SIZE + PADDING)*10, "Log", loadImage(dataPath("images/icons/32/log.png")));
  TOOL_RELOAD         = new EMCCVTool(MARGIN*3 + (EMCCVTool.ICON_SIZE + PADDING)*11, "Reload", loadImage(dataPath("images/icons/32/reload.png")));
  TOOL_DASH           = new EMCCVTool(MARGIN*3 + (EMCCVTool.ICON_SIZE + PADDING)*12, "Dash", loadImage(dataPath("images/icons/32/close.png")));
  
  TOOL_LOG.active = false;
  if (logging && logInfo) log.append(getTime() + "[SETUP] Tools loaded.\n");
}
public void loadThumbnails() {
  fill(GUI_ACTIVE);
  rect(0, IMG_Y, width, height-IMG_Y);

  PImage img;
  int x = MARGIN;
  int y = IMG_Y + MARGIN;

  if (logging && logActions) log.append(getTime() + "[DRAW][LOAD] Number of thumbs in list: " + EMCCVThumb.thumbs.size() + "\n");

  for (int i = 0; i < EMCCVThumb.thumbs.size(); i++) {

    EMCCVThumb thumb = EMCCVThumb.thumbs.get(i);
    thumb.x = x;
    thumb.y = y;
    img = loadImage(sketchPath(DIR_ENTITIES + thumb.SRC + ".png"));

    if (logging && logGui) log.append(getTime() + "[DRAW][LOAD] drawing: SRC = " + thumb.SRC + "\n");
    if (logging && logGui) log.append(getTime() + "[DRAW][LOAD]            x = " + thumb.x + "\n");
    if (logging && logGui) log.append(getTime() + "[DRAW][LOAD]            y = " + thumb.y + "\n");

    fill(GUI_VIEW);
    rect(x, y, EMCCVThumb.W, EMCCVThumb.H);
    noTint();
    image(img, x, y);
    //fill(GUI_TEXT);
    //textAlign(CENTER);
    //text(thumb.SRC, x, y, EMCCVThumb.W, TEXTSIZE*2);

    x = x + EMCCVThumb.W + 2;

    if (x > width-EMCCVThumb.W-MARGIN) {
      if (logging && logGui) log.append(getTime() + "[DRAW][LOAD] NEW ROW\n");
      x = MARGIN;
      y += (EMCCVThumb.H + 2);
    }
  }
}

public StringBuilder printLog(StringBuilder log, String directoy) {
  if (log.length() > 0) {
        
    String filename = "logs/"+ logDirectory + "/"  + directoy + getDateTime() + ".txt";
    
    PrintWriter output = createWriter(filename);
    output.println(log);
    output.flush(); // Write the remaining data
    output.close(); // Finish the file
    //println(log);
  }
  return new StringBuilder();
}

public static void setLogDirectory() {
  SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMdd");
  
  if (activeModule.id != null && !activeModule.id.isEmpty()) {
    logDirectory = activeModule.id + "/" + sdf.format(new Date());
  } else {
    logDirectory = "_dashboard/" + sdf.format(new Date());
  }
}
public abstract static class EMCCVElement {  
  public String parent = "";
  public String xmlParent;
  public String colormaster;
  public int colour;
  public boolean colorizable = true;
  public String blendmode = "blend";
  public boolean flipped = false;
  public int x = 0;
  public int y = 0;
  public int w = 0;
  public int h = 0;
  public char randomize; // shortcut
  public float layer;

  public abstract boolean isActive();

  public EMCCVElement getColormaster() {
    return elements.get(this.colormaster);
  }

  public EMCCVMenuElement getXMLParent() {
    return elements.get(this.xmlParent);
  }
  
  public int getColour() {
    if (this instanceof EMCCVSet || this instanceof EMCCVAspect) {
      return colour;
    } else {
      return elements.get(xmlParent).getColour();
    }
  }
}
public static class EMCCVModule extends EMCCVMenuElement {
  
  String id = null;
  String title = "Extendable Modular Creative Concept Visualizer";
  String authorid = "unknown";
  String version = "";
  String sDesc = "";
  String lDesc = "";
  PImage logo;
  boolean randomizecolor = true;
  
  public boolean selected = false;
  public ArrayList<String> sets = new ArrayList<String>();

  public EMCCVSet activeSet;

  @Override
    public void toggleState() {
    selected = !selected;
  }

  @Override
  public boolean isActive() {
    if (logging && logInfo) log.append(getTime() + "[DRAW][IS ACTIVE][MODULE] id - " + this.id + "\n");
    
    if (this.disabled) return false;
    
    if (this.checked == 0) {
      boolean bResult = !this.disabled;

      if (bResult) {
        checked = 1; // visible
      } else {
        checked = -1; // not visible
      }
    }
    return this.checked == 1;
  }

  @Override
    public void randomize() {
    for (String setID : this.sets) {
      EMCCVSet set = (EMCCVSet) elements.get(setID);
      if (set != null) {
        set.randomize();
      }
    }
  }
}
public abstract static class EMCCVMenuElement extends EMCCVElement {  
  public String id;
  public String title = "unknown title";
  public boolean incognito = false;
  public boolean disabled = false;
  public int checked = 0; // 0 = not checked; 1 = active; -1 = inactive

  // shortcut for randomizing content
  public String randomize;

  public abstract void toggleState();
  public abstract void randomize();

  public boolean isDisabled() {
    return this.disabled || !checkParents(this.parent);
  }
}


public static class EMCCVThumb {
  
  public static final int W = 152;
  public static final int H = 150;
  public int x;
  public int y;
  public final String SRC;
  
  static ArrayList<EMCCVThumb> thumbs = new ArrayList<EMCCVThumb>();
  
  public EMCCVThumb(final String SRC) {
    this.SRC = SRC;
    
    if (logging && logActions) log.append(getTime() + "Thumbnail created: SRC = " + SRC + "\n");
  }
  
  public boolean mouseOver(int x, int y) {
    if (x >= this.x && x <= this.x + W && 
        y >= this.y && y <= this.y + H) {
      return true;
    } else {
      return false;
    }  
  }
}



public static class EMCCVButton {
  public final int X;
  public final int Y;
  public final int W;
  public final int H;
  public final String type;
  public final EMCCVMenuElement ELEMENT;

  // CHECKBOX, RADIOBUTTONS?
  public EMCCVButton(final int X, final int Y, final int W, final int H, EMCCVMenuElement ELEMENT) {
    this.X = X;
    this.Y = Y;
    this.W = W;
    this.H = H;
    this.type = "toggle";
    this.ELEMENT = ELEMENT;
  }

  // COG, CP, LINK
  public EMCCVButton(final int X, final int Y, final int W, final int H, String type, EMCCVMenuElement ELEMENT) {
    this.X = X;
    this.Y = Y;
    this.W = W;
    this.H = H;
    this.type = type;
    this.ELEMENT = ELEMENT;
  }
  
  public boolean isMouseOver(int x, int y) {
    if (x >= this.X && x <= this.X + this.W && y >= this.Y && y <= this.Y + this.H) {
      return true;
    }
    return false;
  }
}
public static class EMCCVSet extends EMCCVMenuElement {
  public int buttonwidth = 70;
  public int probability = 75;
  public boolean visible = true;
  public boolean selected = false;
  public boolean alwayson = false;
  public boolean randomizecolor = true;
  public ArrayList<String> aspects = new ArrayList<String>();

  public EMCCVAspect activeAspect;

  // SCROLLING
  public int position  = 0; // for scrolling
  public int offset    = 0;   // for scrolling
  public int capacity  = 0;  // updated while building aspect menu. How many items fit onto the screen?
  public boolean scrollable = false;
  /*public int scrollbar_h = 0;
  public int handle_h    = 0;
  public int handle_y    = IMG_Y;*/

  // TOOLS
  public boolean colorlink  = false;
  public boolean colorchain = false;
  public boolean collapse  = true;

  @Override
    public void toggleState() {
    selected = !selected;
  }

  @Override
    public boolean isActive() {
    if (logging && logInfo) log.append(getTime() + "[DRAW][IS ACTIVE][SET] id - " + this.id + "\n");

    if (this.disabled) return false;

    if (this.checked == 0) {
      boolean bResult = true;

      if (this.parent != null && !this.parent.equals("") && !this.parent.contains(" " + this.id + " ")) {      
        bResult = checkParents(this.parent);
      }

      bResult = bResult && !this.disabled;

      if (bResult) {
        checked = 1; // visible
      } else {
        checked = -1; // not visible
      }
    }
    return this.checked == 1;
  }

  @Override
    public void randomize() {
    for (String aspectID : this.aspects) {
      EMCCVAspect aspect = (EMCCVAspect) elements.get(aspectID);

      if (aspect != null) {
        aspect.randomize();

        int r = random.nextInt(100);
        if (aspect.alwayson || r <= aspect.probability) {
          aspect.visible = true;
        } else {
          aspect.visible = false;
        }
      }
    }
  }

  public void drawAspectMenu(PApplet pa) {
    int y = ASPECTMENU_Y; // first position

    this.capacity = (pa.height - ASPECTMENU_Y) / ASPECTMENU_D;

    // DRAWING ASPECT MENU
    for (int i = 0; i < aspects.size(); i++) {
      
      // only draw the element if it's not scrolled off the screen
      if (i >= this.offset) {
        String aspectID = aspects.get(i);
        EMCCVAspect aspect = (EMCCVAspect) elements.get(aspectID);

        if (aspect != null && !aspect.incognito && !(aspect.isDisabled() && TOOL_COLLAPSE.active)) {

          // CHECKBOX
          if (aspect.isDisabled()) {
            pa.fill(GUI_DISABLED);
          } else {
            pa.fill(GUI_ENABLED);
          }
          pa.rect(ASPECTMENU_X, y, CONTROLSIZE, CONTROLSIZE, MENURADIUS); 

          // CHECKMARK
          if (aspect.visible) {
            if (aspect.isDisabled() || aspect.alwayson) {
              pa.tint(GUI_SELECTED_DIS);
            } else {
              pa.tint(GUI_SELECTED);
            }
            pa.image(ICON_CHECK, ASPECTMENU_X, y);
          }

          // CHECKBOX into buttonlist
          EMCCVButton checkbox = new EMCCVButton(ASPECTMENU_X, y, CONTROLSIZE, CONTROLSIZE, aspect);
          buttonList.add(checkbox);

          // COLOR CHAINLINK
          if (TOOL_COLORCHAIN.active && aspect.colorizable) {
            if (aspect.colorchain) {
              pa.tint(EMCCVTool.ICON_ACTIVE);
            } else {
              pa.tint(GUI_SELECTED);
            }
            pa.image(ICON_LINK, LINK_X, y);

            // BUTTON (LINK) into buttonlist
            EMCCVButton link = new EMCCVButton(LINK_X, y, CONTROLSIZE, CONTROLSIZE, "link", aspect);
            buttonList.add(link);
          }

          // COG WHEEL
          if (aspect.getNumberOfChildren() > 1) {
            if (aspect.isDisabled()) {
              pa.tint(GUI_SELECTED_DIS);
            } else {
              pa.tint(GUI_SELECTED);
            }
            int tempX = LINK_X; // position depends on whether linklock is activated.
            if (TOOL_COLORCHAIN.active) {
              tempX = COG_X;
            }
            pa.image(ICON_SETTINGS, tempX, y);

            // BUTTON (COG) into buttonlist
            EMCCVButton cog = new EMCCVButton(tempX, y, CONTROLSIZE, CONTROLSIZE, "cog", aspect);
            buttonList.add(cog);
          }

          // TITLE
          if (aspect.isDisabled()) {
            pa.fill(GUI_TEXT_DIS);
          } else {
            pa.fill(GUI_TEXT);
          }
          pa.text(aspect.title, getAspectTitleX(), y-2, getAspectTitleW(), CONTROLSIZE);

          // COLORPICKER
          if (aspect.colorizable && (aspect.colormaster == null || aspect.colormaster.equals(""))) {
            pa.fill(aspect.colour);
            pa.stroke(pa.hue(aspect.colour), pa.saturation(aspect.colour), pa.brightness(aspect.colour)+15);
            pa.rect(CP_X, y, CONTROLSIZE-2, CONTROLSIZE-2, MENURADIUS);

            // COLORPICKER into buttonlist
            EMCCVButton cp = new EMCCVButton(CP_X, y, CONTROLSIZE, CONTROLSIZE, "cp", aspect);
            buttonList.add(cp);
          }
          pa.noStroke();

          y += ASPECTMENU_D;

          // STYLES
          if (aspect == this.activeAspect && aspect.getNumberOfChildren() > 1) {
            for (Map.Entry<String, EMCCVStylegroup> entry : aspect.stylegroups.entrySet()) {

              // separator
              pa.fill(0, 0, 100, 50);
              pa.rect(0, y-(ASPECTMENU_D-CONTROLSIZE)/2, ASPECTMENU_W, 1);
              y++;

              for (Map.Entry<String, EMCCVStyle> style : entry.getValue().styles.entrySet()) {

                // lighter background color
                pa.fill(0, 0, 255, 15);
                pa.rect(0, y-(ASPECTMENU_D-CONTROLSIZE)/2, ASPECTMENU_W, ASPECTMENU_D);

                EMCCVStyle emccvStyle = (EMCCVStyle) style.getValue();
                if (emccvStyle.isDisabled()) {
                  pa.fill(GUI_DISABLED);
                } else {
                  pa.fill(GUI_ENABLED);
                }
                pa.ellipse(getAspectTitleX()-19, y, CONTROLSIZE, CONTROLSIZE);

                if (emccvStyle.selected) {
                  if (emccvStyle.isDisabled()) {
                    pa.fill(GUI_SELECTED_DIS);
                  } else {
                    pa.fill(GUI_SELECTED);
                  }
                  int size = CONTROLSIZE-8;
                  pa.ellipse(getAspectTitleX()-19+((CONTROLSIZE-size)/2), y+((CONTROLSIZE-size)/2), size, size);
                }

                // TITLE
                if (emccvStyle.isDisabled()) {
                  pa.fill(GUI_TEXT_DIS);
                } else {
                  pa.fill(GUI_TEXT);
                }
                if (emccvStyle.title != null) {
                  pa.text(emccvStyle.title, getAspectTitleX(), y-2, getAspectTitleW(), CONTROLSIZE);
                } else {
                  pa.text("unknown", getAspectTitleX(), y-2, getAspectTitleW(), CONTROLSIZE);
                }


                // BUTTON into buttonlist
                EMCCVButton stylebutton = new EMCCVButton(getAspectTitleX()-19, y, CONTROLSIZE, CONTROLSIZE, emccvStyle);
                buttonList.add(stylebutton);

                y += ASPECTMENU_D;
              }
            }
            // separator
            pa.fill(0, 0, 100, 50);
            pa.rect(0, y-(ASPECTMENU_D-CONTROLSIZE)/2, ASPECTMENU_W, 1);
            y++;
          }
        }
      }
    }
    
    

    // SCROLLING SETUP
    if (this.getNumberOfPositions() <= this.capacity) {
      this.offset = 0;
      this.scrollable = false;
    } else {
      this.scrollable = true;
      pa.fill(GUI_ENABLED);
      pa.rect(ASPECTMENU_W-32, IMG_Y, 32, 32*4);
      pa.tint(GUI_SELECTED);
      pa.image(ICON_FIRST, ASPECTMENU_W-28, IMG_Y+4);
      pa.image(ICON_UP,    ASPECTMENU_W-28, IMG_Y+36);
      pa.image(ICON_DOWN,  ASPECTMENU_W-28, IMG_Y+68);
      pa.image(ICON_LAST,  ASPECTMENU_W-28, IMG_Y+100);
    }
  }
  
  public int getNumberOfPositions() {
    int result = 0;
    
    // count available aspects (on and offscreen)
    for (String aspectid : this.aspects) {
      if (elements.get(aspectid) != null && (!this.collapse || !elements.get(aspectid).isDisabled())) {
        result++;
      }
    }
    
    if (this.activeAspect != null) {
      result = result + this.activeAspect.getNumberOfChildren();
    }
    return result;
  }
}
public static class EMCCVAspect extends EMCCVMenuElement {
  public static int WIDTH = 280;

  public int probability = 75;
  public boolean visible = false;
  public String style;
  public boolean alwayson = false;
  public boolean randomizecolor = true;
  public Map<String, EMCCVStylegroup> stylegroups = new LinkedHashMap<String, EMCCVStylegroup>();
  
  public boolean colorchain = false;

  @Override
  public boolean isActive() {
    if (logging && logInfo) log.append(getTime() + "[DRAW][IS ACTIVE][ASPECT] id - " + this.id + "\n");
    
    if (this.disabled || (!this.visible && !this.alwayson)) return false;
    
    if (this.checked == 0) {
      boolean bResult = true;

      if (!alwayson) {
        EMCCVMenuElement xmlParent = this.getXMLParent();
        if (xmlParent != null) {
          if (logging && logInfo) log.append(getTime() + "[DRAW][IS ACTIVE][ASPECT] xmlParent.id - " + xmlParent.id + "\n");
          bResult = xmlParent.isActive();
        }
      } else if (logging && logInfo) {
        log.append(getTime() + "[DRAW][IS ACTIVE][ASPECT] " + this.id + " is ALWAYSON.\n");
      }

      if (bResult && this.parent != null && !this.parent.equals("")) {      
        bResult = bResult && checkParents(this.parent);
      }

      if (bResult) {
        checked = 1; // visible
      } else {
        checked = -1; // not visible
      }
    }
    return this.checked == 1;
  }

  @Override
    public void toggleState() {
    if (this.alwayson) {
      visible = true;
    } else {
      visible = !visible;
    }
  }

  public int getNumberOfChildren() {
    int numberofstyles = 0;

    for (Map.Entry<String, EMCCVStylegroup> entry : this.stylegroups.entrySet()) {
      numberofstyles += entry.getValue().styles.size();
    }

    return numberofstyles;
  }

  @Override
    public boolean isDisabled() {
    return super.isDisabled() || !this.getXMLParent().isActive();
  }

  @Override
    public void randomize() {
    for (Map.Entry<String, EMCCVStylegroup> entry : this.stylegroups.entrySet()) {
      boolean success = false;
      while (!success) {
        success = entry.getValue().randomize();
      }
    }
  }
}
public static class EMCCVStylegroup {  
  public String title = "default";
  public Map<String, EMCCVStyle> styles = new LinkedHashMap<String, EMCCVStyle>();

  public EMCCVStylegroup(String title) {
    this.title = title;
  }

  public void resetStyles() {
    for (Map.Entry<String, EMCCVStyle> entry : styles.entrySet()) {
      entry.getValue().selected = false;
    }
  }


  public boolean randomize() { 
    resetStyles();
    boolean success = false;
    int index = 0;

    int repeat = 0;

    while (!success) {
      index = 0; // the index of the style in this stylegroup
      int r = random.nextInt(styles.size());

      for (Map.Entry<String, EMCCVStyle> entry : styles.entrySet()) {
        if (r == index && checkParents(entry.getValue().parent)) {
          entry.getValue().selected = true;
          success = true;
          if (logging && repeat > 3) {
            log.append(getTime() + "[RANDMONIZATION PARENT WARNING] ATTEMPT #" + repeat + "; stylegroup '" + this.title + "' for '" + entry.getValue().xmlParent + "'\n");
          }
          break;
        } else {
          // this style can not be selected. Try the next one.
          index++;
          if (index == styles.size()) {
            resetStyles();
            repeat++;
            if (repeat > 10) {
              if (logging) log.append(getTime() + "[RANDMONIZATION PARENT ERROR]   Given up after 10 attempts: stylegroup '" + this.title + "' for '" + entry.getValue().xmlParent + "'\n");
              if (logging) log.append(getTime() + "|||||||||||||||||||||||||||||   Restarting randomization process for ASPECT '" + entry.getValue().xmlParent + "'\n");
              return success;
            }
          }
        }
      }
    }
    return success;
  }
}
public static class EMCCVStyle extends EMCCVMenuElement {
  public static int WIDTH = 280;

  public boolean selected = false;
  public String stylegroup = "default";
  public ArrayList<EMCCVFile> files = new ArrayList<EMCCVFile>();


  @Override
  public boolean isActive() {
    if (this.disabled || !this.selected) return false;
    if (this.checked == 0) {
      boolean bResult = true;
      
      EMCCVMenuElement xmlParent = this.getXMLParent();
      if (xmlParent != null) {
        if (logging && logInfo) log.append(getTime() + "[DRAW][IS ACTIVE][STYLE] id - " + this.id + "\n");
        if (logging && logInfo) log.append(getTime() + "[DRAW][IS ACTIVE][STYLE] xmlParent.id - " + xmlParent.id + "\n");
        bResult = xmlParent.isActive();
      }
  
      if (bResult && this.parent != null && !this.parent.equals("")) {      
        bResult = bResult && checkParents(this.parent);
      }
  
      bResult = bResult && !this.disabled && this.selected;

      if (bResult) {
        checked = 1;
      } else {
        checked = -1;
      }
    }
    return this.checked == 1;
  }

  @Override
    public void toggleState() {
    selected = !selected;
  }
  
  @Override
  public void randomize() {
    // do nothing;
  }
}
public static class EMCCVFile extends EMCCVElement {

  public String src;
  
  


  @Override
  public boolean isActive() {
    boolean bResult = true;

    EMCCVMenuElement xmlParent = this.getXMLParent();
    if (xmlParent != null) {
      if (logging && logInfo) log.append(getTime() + "[DRAW][IS ACTIVE][FILE] src: " + this.src + "\n");
      if (logging && logInfo) log.append(getTime() + "[DRAW][IS ACTIVE][FILE] xmlParent: " + this.xmlParent + "\n");
      bResult = xmlParent.isActive();
    }

    if (bResult && this.parent != null && !this.parent.equals("")) {
      bResult = bResult && checkParents(this.parent);
    }
    
    return bResult;
  }
}
// GUI CONSTANTS
final static int PADDING      = 10;
final static int MARGIN       = 20;
final static int MENURADIUS   = 4;
final static int TEXTSIZE     = 13;
final static int CONTROLSIZE  = 16;

// TOOLS
public static EMCCVTool TOOL_LOAD;
public static EMCCVTool TOOL_SAVE;
public static EMCCVTool TOOL_CAPTURE;
public static EMCCVTool TOOL_COLORLINK;
public static EMCCVTool TOOL_COLORCHAIN;
public static EMCCVTool TOOL_COLLAPSE;
public static EMCCVTool TOOL_GLOBALMODE;
public static EMCCVTool TOOL_RANDOMIZE;
public static EMCCVTool TOOL_LOG;
public static EMCCVTool TOOL_RELOAD;
public static EMCCVTool TOOL_DASH;
public static EMCCVTool TOOL_COLORSHIFT;
public static EMCCVTool TOOL_RANDOMIZECOLOR;

// IMAGES
public static PImage IMG_LOGO;
public static PImage IMG_CREDITS;
public static PImage IMG_GRADIENT_HUE;
public static PImage IMG_GRADIENT_SAT;
public static PImage IMG_GRADIENT_BRI;
public static PImage ICON_CHECK;
public static PImage ICON_LINK;
public static PImage ICON_SETTINGS;

public static PImage ICON_FIRST;
public static PImage ICON_UP;
public static PImage ICON_DOWN;
public static PImage ICON_LAST;

// SETS
final static int MENUSET_W = 70;
final static int MENUSET_H = 19;
final static int MENUSET_Y = EMCCVTool.TOOLBAR_Y + EMCCVTool.TOOLBAR_H + MARGIN;
final static int MENUSET_D = 2;
final static int MENUSET_P = 10;

// ASPECTS
final static int ASPECTMENU_W = 285;
final static int ASPECTMENU_X = MARGIN;
final static int ASPECTMENU_Y = MENUSET_Y + MENUSET_H + ASPECTMENU_X;
final static int ASPECTMENU_D = 22;

// COLOR PICKER
final static int GRADIENT_W   = 65;
final static int GRADIENT_H   = 255;
final static int GRADIENT_H_X = ASPECTMENU_X;
final static int GRADIENT_S_X = ASPECTMENU_X + GRADIENT_W + PADDING;
final static int GRADIENT_B_X = ASPECTMENU_X + GRADIENT_W*2 + PADDING*2;
final static int GRADIENT_Y   = ASPECTMENU_Y + GRADIENT_W + MARGIN;

final static int SWATCH_X     = GRADIENT_H_X;
final static int SWATCH_Y     = GRADIENT_Y + GRADIENT_H + 80;
static ArrayList colorlist = new ArrayList();

final static int CP_X         = ASPECTMENU_X+19;
final static int LINK_X       = ASPECTMENU_X+38;
final static int COG_X        = ASPECTMENU_X+57;

public static int getAspectTitleX() {
  if (TOOL_COLORCHAIN.active) {
    return COG_X+19;
  }
  return LINK_X+19;
}
public static int getAspectTitleW() {
  return ASPECTMENU_W - getAspectTitleX();
}

final static int IMG_X = ASPECTMENU_W;
final static int IMG_Y = MENUSET_H + MENUSET_Y;

int currentWidth = width;
int currentHeight = height;
int port_W = displayWidth;
int port_H = displayHeight;
float factor = 1.0f;
int view_W = displayWidth;
int view_H = displayHeight;


// COLORS
final static int GUI_ENABLED      = 0xFF646464; // color(100, 100, 100);
final static int GUI_DISABLED     = 0x50646464; // color(100, 100, 100, 80);
final static int GUI_ACTIVE       = 0xFF808080; // color(128, 128, 128);
final static int GUI_VIEW         = 0xFF878787; // color(135, 135, 135);
final static int GUI_SELECTED     = 0xFFB4B4B4; // color(180, 180, 180);
final static int GUI_SELECTED_DIS = 0x50B4B4B4; // color(180, 180, 180, 80);
final static int GUI_BG           = 0xFF505050; // color(80, 80, 80);
final static int GUI_TEXT         = 0xFFDCDCDC; // color(220, 220, 220);
final static int GUI_TEXT_DIS     = 0x50DCDCDC; // color(220, 220, 220, 80);

// GRADIENT
int Y_AXIS = 1;
int X_AXIS = 2;

// GUI STATES
static ArrayList<EMCCVSet> setList = new ArrayList<EMCCVSet>();
static ArrayList<EMCCVButton> buttonList = new ArrayList<EMCCVButton>();
static EMCCVAspect activeAspect;

static boolean redrawModel = true;
static boolean refreshModel = false;
static boolean redrawSetMenu = true;
static boolean redrawToolbar = true;
static boolean redrawAspectMenu = true;
static int tempColor = 0;
static boolean bTempColor = false;


//////////////////////////////////////////////////////////////////////////////////////////////// UPDATE VIEW

public void updateView() {
  if (port_W != 0) {
    port_W = width - IMG_X;
  } else port_W = displayWidth - IMG_X - 20;

  if (port_H != 0) {
    port_H = height - IMG_Y;
  } else port_H = displayHeight - IMG_Y - 120;

  view_W = floor(activeModule.w * factor);
  view_H = floor(activeModule.h * factor);

  if (logging && logGui) {
    log.append(getTime() + "====================================\n");
    log.append(getTime() + "[UPDATEVIEW BEFORE] width: "  + width  + "\n");
    log.append(getTime() + "[UPDATEVIEW BEFORE] height: " + height + "\n");
    log.append(getTime() + "[UPDATEVIEW BEFORE] port_W: " + port_W + "\n");
    log.append(getTime() + "[UPDATEVIEW BEFORE] port_H: " + port_H + "\n");
    log.append(getTime() + "[UPDATEVIEW BEFORE] view_W: " + view_W + "\n");
    log.append(getTime() + "[UPDATEVIEW BEFORE] view_H: " + view_H + "\n");
    log.append(getTime() + "[UPDATEVIEW BEFORE] factor: " + factor + "\n");
    log.append(getTime() + "[UPDATEVIEW BEFORE] factor_W: " + (float) port_W / (float) activeModule.w + "\n");
    log.append(getTime() + "[UPDATEVIEW BEFORE] factor_H: " + (float) port_H / (float) activeModule.h + "\n");
  }
  float factor_W = (float) port_W / (float) activeModule.w;
  float factor_H = (float) port_H / (float) activeModule.h;

  if (factor_W < factor_H) {
    factor = factor_W;
  } else {
    factor = factor_H;
  }

  if (factor > 1) {
    factor = 1;
  }

  view_W = floor(activeModule.w * factor);
  view_H = floor(activeModule.h * factor);

  if (logging && logGui) {
    log.append(getTime() + "[UPDATEVIEW AFTER] view_W: " + view_W + "\n");
    log.append(getTime() + "[UPDATEVIEW AFTER] view_H: " + view_H + "\n");
    log.append(getTime() + "[UPDATEVIEW AFTER] factor_W: " + factor_W + "\n");
    log.append(getTime() + "[UPDATEVIEW AFTER] factor_H: " + factor_H + "\n");
    log.append(getTime() + "[UPDATEVIEW AFTER] factor: " + factor + "\n");
    log.append(getTime() + "====================================\n");
  }
}


//////////////////////////////////////////////////////////////////////////////////////////////// DRAW SET MENU

public void drawMenuSets() {
  fill(GUI_BG);
  rect(0, 0, width, MENUSET_H + MENUSET_Y);
  textAlign(CENTER, CENTER);

  int x = MENUSET_D;
  for (EMCCVSet set : setList) {

    if (!set.incognito) {
      if (set.selected) {
        fill(GUI_ACTIVE);
        activeModule.activeSet = set;
      } else {
        fill(GUI_ENABLED);
      }

      rect(x, MENUSET_Y, set.buttonwidth, MENUSET_H, MENURADIUS, MENURADIUS, 0, 0);
      fill(GUI_TEXT);
      text(set.title, x, MENUSET_Y-2, set.buttonwidth, MENUSET_H);

      EMCCVButton button = new EMCCVButton(x, MENUSET_Y, set.buttonwidth, MENUSET_H, set);
      buttonList.add(button);

      x += set.buttonwidth + MENUSET_D;
    }
  }
  noTint();
  image(IMG_LOGO, width-PADDING-254, (EMCCVTool.TOOLBAR_Y-40)/2);

  textAlign(LEFT, CENTER);
  fill(GUI_TEXT);
  textSize(16);
  text(activeModule.title, PADDING, 0, width-PADDING-254-PADDING*2, EMCCVTool.TOOLBAR_Y);
  textSize(TEXTSIZE);
}

//////////////////////////////////////////////////////////////////////////////////////////////// DRAW TOOLBAR
public void drawToolbar() {  
  fill(255, 0, EMCCVTool.TOOLBAR_BG);
  rect(0, EMCCVTool.TOOLBAR_Y, width, EMCCVTool.TOOLBAR_H );

  for (EMCCVTool tool : EMCCVTool.tools) {
    tint(255, tool.getColor());
    image(tool.ICON, tool.X, EMCCVTool.Y);
  }
}


//////////////////////////////////////////////////////////////////////////////////////////////// DRAW ASPECT MENU

public void drawAspectMenu() {

  resetButtonList();

  // ASPECT MENU SETUP
  fill(GUI_ACTIVE);
  rect(0, IMG_Y, ASPECTMENU_W, height - IMG_Y); // background
  textAlign(LEFT, CENTER);
  //int o = activeModule.activeSet.offset; // offset for scrolling  

  // DRAWING THE MENU
  if(setList.isEmpty()) {
    log.append(getTime() + " [DRAWING ASPECT MENU] ERROR: The active MODULE seems to have no sets.\n");
  } else if (activeModule.activeSet == null) {
    activeModule.activeSet = (EMCCVSet) elements.get(setList.get(0).title);
    log.append(getTime() + " [DRAWING ASPECT MENU] ERROR: The active MODULE was null. Something is wrong here...\n");
  }
    if (setList.get(0) == activeModule.activeSet) {
      
      drawInfoTab();
      
    } else {
      activeModule.activeSet.drawAspectMenu(this);
    }
    noTint();
}


//////////////////////////////////////////////////////////////////////////////////////////////// DRAW MODULE MENU

public void drawModuleMenu() {
  fill(GUI_ACTIVE);
  rect(0, IMG_Y, width, height-IMG_Y);

  int w = 100;
  int h = 40;
  int x = MARGIN;
  int y = IMG_Y+MARGIN;
  int xcount = floor((width-MARGIN*2) / (w+PADDING));

  loadModuleIDs();
  if (moduleids.size() > 0) {

    for (int i = 0; i < moduleids.size(); i++) {
      x = MARGIN       + ((i * (w + PADDING)) % (xcount * (w+PADDING)));
      if ((i+1) % (xcount+1) == 0) {
        y = y + h + PADDING;
      }
      //y = IMG_Y+MARGIN + ((i * (h + PADDING)) / (ycount * (h+PADDING)));

      fill(GUI_ENABLED);
      rect(x, y, w, h, 3);
      fill(GUI_TEXT);
      text(moduleids.get(i), x+PADDING, y, w-2*PADDING, h);

      // MODULE into elements list
      EMCCVModule module = new EMCCVModule();
      module.id = moduleids.get(i);
      elements.put(module.id, module);

      // BUTTON into buttonlist
      EMCCVButton modulebutton = new EMCCVButton(x, y, w, h, "module", module);
      buttonList.add(modulebutton);
    }
  } else {
    fill(GUI_ENABLED);
    rect(x, y, w, h, 3);
    fill(GUI_TEXT);
    text("No modules found.", x+PADDING, y, w-2*PADDING, h);
  }
}
public void resetButtonList() {
  // remove all Aspect and Style buttons from the list
  ArrayList<EMCCVButton> deleteList = new ArrayList<EMCCVButton>();
  for (EMCCVButton button : buttonList) {
    if (!(button.ELEMENT instanceof EMCCVSet)) {
      deleteList.add(button);
    }
  }
  for (EMCCVButton button : deleteList) {
    buttonList.remove(button);
  }
  deleteList.clear();
}

public void drawInfoTab() {
    int y = ASPECTMENU_Y;
    noTint();
    try {
      image(activeModule.logo, ASPECTMENU_X, y, 245, 60);
    } 
    catch (Exception e) {
      log.append(getTime() + "[ERROR] No Module logo found.");
    }

    StringBuilder sb = new StringBuilder("EMCCV Prototype " + EMCCV_VERSION + "\n");

    if (activeModule.id != null) {
      sb.append("activeModule id: ").append(activeModule.id + " v" + activeModule.version + "\n");
    }
    if (activeModule.authorid != null) {
      sb.append("author id: ").append(activeModule.authorid + "\n");
    }
    if (activeModule.sDesc != null) {
      sb.append("\n" + activeModule.sDesc);
    }
    if (randomKeys.size() > 0) {
      sb.append("\n\nRandomization shortcuts: \n");
      for (Map.Entry<String, EMCCVMenuElement> entry : randomKeys.entrySet()) {
        sb.append(entry.getKey() + ": ");
        sb.append(entry.getValue().title + "\n");
      }
    }
    textAlign(LEFT, TOP);
    fill(GUI_TEXT);
    text(sb.toString(), ASPECTMENU_X, y+60+MARGIN, 256, 500);
}
public void drawColorPicker() {
  
  // Background
  fill(GUI_ACTIVE);
  rect(0, MENUSET_Y + MENUSET_H, ASPECTMENU_W, height - MENUSET_Y - MENUSET_H);
  
  // Title
  textAlign(LEFT, CENTER);
  fill(GUI_TEXT);
  text("COLOR PICKER: " + activeAspect.title, ASPECTMENU_X, ASPECTMENU_Y);
  text("0", GRADIENT_B_X + GRADIENT_W + PADDING, GRADIENT_Y);
  text("255", GRADIENT_B_X + GRADIENT_W + PADDING, GRADIENT_Y + GRADIENT_H);
  
  textAlign(CENTER, CENTER);
  text("HUE", GRADIENT_H_X, GRADIENT_Y + GRADIENT_H + PADDING, GRADIENT_W, MARGIN);
  text("SAT", GRADIENT_S_X, GRADIENT_Y + GRADIENT_H + PADDING, GRADIENT_W, MARGIN);
  text("BRI", GRADIENT_B_X, GRADIENT_Y + GRADIENT_H + PADDING, GRADIENT_W, MARGIN);
  textSize(26); 
  text("" + floor(hue(activeAspect.colour)), GRADIENT_H_X, GRADIENT_Y + GRADIENT_H + PADDING*2, GRADIENT_W, MARGIN*2);
  text("" + floor(saturation(activeAspect.colour)), GRADIENT_S_X, GRADIENT_Y + GRADIENT_H + PADDING*2, GRADIENT_W, MARGIN*2);
  text("" + floor(brightness(activeAspect.colour)), GRADIENT_B_X, GRADIENT_Y + GRADIENT_H + PADDING*2, GRADIENT_W, MARGIN*2);
  textSize(TEXTSIZE);
  
  // Active Colour
  fill(activeAspect.colour);
  rect(GRADIENT_H_X, ASPECTMENU_Y + 20, GRADIENT_W*3 + PADDING*2, GRADIENT_W-PADDING);
  
  // Original Colour
  if (!bTempColor) {
    tempColor = activeAspect.colour;
    bTempColor = true;
  }
  
  fill(tempColor);
  stroke(GUI_ACTIVE);
  rect(GRADIENT_H_X-1, ASPECTMENU_Y + 19, 20, 20);
  noStroke();
  
  // HUE
  fill(0, 0, brightness(activeAspect.colour));
  rect(GRADIENT_H_X, GRADIENT_Y, GRADIENT_W, GRADIENT_H);
  tint(0, 0, brightness(activeAspect.colour), saturation(activeAspect.colour));
  image(IMG_GRADIENT_HUE, GRADIENT_H_X, GRADIENT_Y, GRADIENT_W, GRADIENT_H);
  
  // SATURATION
  fill(0, 0, brightness(activeAspect.colour));
  rect(GRADIENT_S_X, GRADIENT_Y, GRADIENT_W, GRADIENT_H);
  tint(hue(activeAspect.colour), 255, brightness(activeAspect.colour));
  image(IMG_GRADIENT_SAT, GRADIENT_S_X, GRADIENT_Y, GRADIENT_W, GRADIENT_H);
  
  // BRIGHTNESS
  tint(hue(activeAspect.colour), saturation(activeAspect.colour), 255);
  image(IMG_GRADIENT_BRI, GRADIENT_B_X, GRADIENT_Y, GRADIENT_W, GRADIENT_H);
  
  // PALETTE
  int palX;
  int palY = SWATCH_Y;
  colorlist.clear();
  for (int i = 0; i < COLOURS_50BRI.size(); i++) {
    int c = color((int) COLOURS_50BRI.get(i));
    colorlist.add(c);
  }
  for (EMCCVFile file : fileLayers) {
    if (!colorlist.contains(file.getColour())) {
      
      colorlist.add(file.getColour());
    }
  }
  for (Object o : colorlist) {
    //if (colorlist.size() == 0 || hue(file.getColour()) > hue((color) colorlist.get(colorlist.size()-1))) {
    //  colorlist.add(file.getColour());
    //} else {
    //  for (int i = 0; i < colorlist.size(); i++) {
    //    if (hue(file.getColour()) > hue((color) colorlist.get(i))) {
    //      colorlist.add(i, file.getColour());
    //      break;
    //    }
    //  }
    //}
    
    fill((int) o);
    palX = SWATCH_X + MARGIN  * (colorlist.indexOf(o) % 10) + 1 * (colorlist.indexOf(o) % 10);
    palY = SWATCH_Y + MARGIN * (colorlist.indexOf(o) / 10) + 1 * (colorlist.indexOf(o) / 10);
    rect(palX, palY, MARGIN, MARGIN); 
  }
  
  
  // MARKERS
  if (saturation(activeAspect.colour) < 60 && brightness(activeAspect.colour) > 230) {
    stroke(0);
  } else {
    stroke(255);
  }
  line(GRADIENT_H_X -1,         GRADIENT_Y + hue(activeAspect.colour),
       GRADIENT_H_X+GRADIENT_W, GRADIENT_Y + hue(activeAspect.colour));
       
  line(GRADIENT_S_X -1,         GRADIENT_Y + saturation(activeAspect.colour),
       GRADIENT_S_X+GRADIENT_W, GRADIENT_Y + saturation(activeAspect.colour));
       
  line(GRADIENT_B_X -1,         GRADIENT_Y + brightness(activeAspect.colour),
       GRADIENT_B_X+GRADIENT_W, GRADIENT_Y + brightness(activeAspect.colour));
  
  noFill();
  noTint();
  noStroke();
}

public void exitCP() {
  mode = PLAY;
  bTempColor = false;
  redrawAspectMenu = true;
  activeAspect = null;
}
public void mouseMoved() {
  cursor(ARROW); // default

  if (mode != DASH) {

    // TOOLBAR
    if (mouseY > EMCCVTool.TOOLBAR_Y && mouseY < EMCCVTool.TOOLBAR_Y + EMCCVTool.TOOLBAR_H) {
      if (mode != SAVE && mode != RELOAD && mode != SAVE && mode != CAPTURE) {
        for (EMCCVTool tool : EMCCVTool.tools) {
          if (tool.mouseOver(mouseX, mouseY)) {
            cursor(HAND);
            break;
          }
        }
      }
    } else if (mode == CP && (mouseOverHue() || mouseOverSat() || mouseOverBri() || mouseOverModel() || mouseOverTempColor() || mouseOverSwatch())) {
      cursor(CROSS);
    } else if (TOOL_LOAD.active) {
      for (EMCCVThumb thumb : EMCCVThumb.thumbs) {
        if (thumb.mouseOver(mouseX, mouseY)) {
          cursor(HAND);
          break;
        }
      }
    } else if (mouseX >= ASPECTMENU_X && mouseX <= getAspectTitleX() || mouseY >= MENUSET_Y && mouseY <= MENUSET_Y + MENUSET_H) {
      for (EMCCVButton button : buttonList) {
        if (button.isMouseOver(mouseX, mouseY)) {
          cursor(HAND);
          break;
        }
      }
    } else if (activeModule.activeSet.scrollable && mouseX >= ASPECTMENU_W-32 && mouseX <= ASPECTMENU_W && mouseY >= IMG_Y && mouseY <= IMG_Y + 32*4) {
      cursor(HAND);
    }
  } else if (TOOL_LOG.mouseOver(mouseX, mouseY)) {
    cursor(HAND);    
  } else if (moduleids.size() > 0) {
    for (EMCCVButton button : buttonList) {
      if (button.isMouseOver(mouseX, mouseY)) {
        cursor(HAND);
        break;
      }
    }
  }
}

public void mouseClicked() {

  if (mode != DASH) {

    if (mouseY >= EMCCVTool.Y && mouseY <= EMCCVTool.Y + EMCCVTool.ICON_SIZE) {
      if (mode != SAVE && mode != RELOAD && mode != SAVE && mode != CAPTURE) {
        reactToToolbarClick(); // TOOLBAR CLICK
      }
    } else if (mode == CP) {
      if (mouseY < IMG_Y) {
        exitCP();
        redraw();
      } else {
        processColorpickerClick();
      }
    } else if (mode == LOAD) {
      if (mouseY > IMG_Y && EMCCVThumb.thumbs.size() > 0) {
        processThumbnailClick();
      } else {
        mode = PLAY;
        redraw();
      }
    } else if (mode == PLAY) {
      if (activeModule.activeSet.scrollable && mouseX >= ASPECTMENU_W-32 && mouseX <= ASPECTMENU_W && mouseY >= IMG_Y && mouseY <= IMG_Y+128) {
        if (mouseY <= IMG_Y+32) {
          activeModule.activeSet.offset = 0;
        } else if (mouseY <= IMG_Y+64) {
          activeModule.activeSet.offset--;
          if (activeModule.activeSet.offset < 0) {
            activeModule.activeSet.offset = 0;
          }
        } else if (mouseY <= IMG_Y+96) {
          activeModule.activeSet.offset++;
        } else if (mouseY <= IMG_Y+128) {
          activeModule.activeSet.offset = activeModule.activeSet.getNumberOfPositions() - ((height - ASPECTMENU_Y) / ASPECTMENU_D);
        }
        redrawAspectMenu = true;
        redraw();

      } else if ((mouseX >= ASPECTMENU_X && mouseX <= COG_X+CONTROLSIZE) || mouseY >= MENUSET_Y && mouseY <= MENUSET_Y + MENUSET_H) {
        processButtonClick();
      }
    }
  } else if (TOOL_LOG.mouseOver(mouseX, mouseY)) {
    if (logging && logActions) log.append(getTime() + " LOG CLICK\n");
    reactToToolbarClick();
  } else if (moduleids.size() > 0) {
    for (EMCCVButton button : buttonList) {
      if (button.isMouseOver(mouseX, mouseY)) {
        EMCCVModule module = (EMCCVModule) button.ELEMENT;
        initXML(module.id);
        mode = PLAY;
        TOOL_DASH.active = false;
        redrawSetMenu    = true;
        redrawAspectMenu = true;
        redrawToolbar    = true;
        redrawModel      = true;
        redraw();
      }
    }
  }

  // RESIZE WHEN NEEDED
  if (currentWidth != width || currentHeight != height) {
    redrawSetMenu = true;
    redrawAspectMenu = true;
    redrawModel = true;
    currentWidth = width;
    currentHeight = height;
    redraw();
  }
}

public void keyPressed() {
  String sKey = String.valueOf(key);

  if (sKey.equals("0") ||
    sKey.equals("1") ||
    sKey.equals("2") ||
    sKey.equals("3") ||
    sKey.equals("4") ||
    sKey.equals("5") ||
    sKey.equals("6") ||
    sKey.equals("7") ||
    sKey.equals("8") ||
    sKey.equals("9")) {

    int index = Integer.valueOf(sKey);

    if (index < setList.size()) {
      for (EMCCVSet set : setList) {
        set.selected = false;
      }

      setList.get(index).selected = true;

      redrawSetMenu = true;
      exitCP();

      redraw();
    }
  } else if (mode != CP && randomKeys.containsKey(sKey)) {

    if (!(randomKeys.get(sKey) instanceof EMCCVAspect) || randomKeys.get(sKey).isActive()) {
      randomKeys.get(sKey).randomize();
      redrawAspectMenu = true;
      redrawModel = true;
      redraw();
    }
  }
}

public void processButtonClick() {
  for (EMCCVButton button : buttonList) {

    if (mouseX >= button.X && mouseX <= button.X + button.W && mouseY >= button.Y && mouseY <= button.Y + button.H) {

      if (button.ELEMENT instanceof EMCCVSet) {
        if (logging && logActions) log.append(getTime() + " CLICKED: " + button.ELEMENT.title + " (SET)\n");

        for (EMCCVSet set : setList) {
          set.selected = false;
        }
        activeModule.activeSet = (EMCCVSet) button.ELEMENT;
        button.ELEMENT.toggleState();

        if (!TOOL_GLOBALMODE.active) { 
          TOOL_COLORLINK.active = activeModule.activeSet.colorlink;
          TOOL_COLORCHAIN.active = activeModule.activeSet.colorchain;
          TOOL_COLLAPSE.active = activeModule.activeSet.collapse;
          redrawToolbar    = true;
        }
        redrawSetMenu    = true;
        redrawAspectMenu = true;
        redrawModel      = true;
        TOOL_LOAD.active = false;
        exitCP();
        redraw();
      } else if (mode != CP && button.ELEMENT instanceof EMCCVAspect) {

        EMCCVAspect aspect = (EMCCVAspect) button.ELEMENT;
        if (logging && logInfo) log.append(getTime() + "[CHECKPARENTS] of " + aspect.title + ": " + checkParents(aspect.parent));

        if (button.type.equals("cp")) {
          if (logging && logActions) log.append(getTime() + " CLICKED: " + button.ELEMENT.title + " (ASPECT CP)\n");
          activeAspect = (EMCCVAspect) button.ELEMENT;
          activeAspect.visible = true;
          mode = CP;
          redraw();
        } else if (button.type.equals("link")) {
          if (logging && logActions) log.append(getTime() + " CLICKED: " + button.ELEMENT.title + " (ASPECT COLORCHAIN)\n");
          aspect.colorchain = !aspect.colorchain;
          redrawAspectMenu = true;
          redraw();
        } else if (button.type.equals("cog")) {
          if (logging && logActions) log.append(getTime() + " CLICKED: " + button.ELEMENT.title + " (ASPECT COG)\n");
          if (aspect == activeModule.activeSet.activeAspect) {
            activeModule.activeSet.activeAspect = null;
            redrawAspectMenu = true;
            redraw();
          } else {
            activeModule.activeSet.activeAspect = aspect;
            aspect.visible = true;
            redrawAspectMenu = true;
            redraw();
          }
        } else if (!aspect.alwayson && !aspect.isDisabled()) {
          if (logging && logActions) log.append(getTime() + " CLICKED: " + button.ELEMENT.title + " (ASPECT VISIBILITY)\n");
          button.ELEMENT.toggleState();
          redrawAspectMenu = true;
          redrawModel = true;
          redraw();
        }
      } else if (mode != CP && button.ELEMENT instanceof EMCCVStyle && !button.ELEMENT.isDisabled()) {
        if (logging && logActions) log.append(getTime() + " CLICKED: " + button.ELEMENT.title + " (STYLE)\n");

        if (button.ELEMENT.getXMLParent() != null) {          
          EMCCVStyle style   = (EMCCVStyle) button.ELEMENT;
          EMCCVAspect aspect = (EMCCVAspect) button.ELEMENT.getXMLParent();

          if (logging && logInfo) log.append(getTime() + "[CHECKPARENTS] " + aspect.title + ": " + checkParents(aspect.xmlParent));

          for (Map.Entry<String, EMCCVStylegroup> entry : aspect.stylegroups.entrySet()) {
            if (entry.getValue().title.equals(style.stylegroup)) {
              entry.getValue().resetStyles();
              break;
            }
          }
          aspect.visible = true;
        }

        button.ELEMENT.toggleState();
        redrawAspectMenu = true;
        redrawModel = true;
        redraw();
      }
    }
  }
}

public void reactToToolbarClick() {
  for (EMCCVTool tool : EMCCVTool.tools) {
    boolean processed = false;
    if (tool.mouseOver(mouseX, mouseY)) {
      tool.active = !tool.active;

      // Set-Tools
      if (tool == TOOL_LOG) {
        if (logging && logActions) log.append(getTime() + " LOG CLICK\n");
        logging = !logging;
      } else if (tool == TOOL_COLORCHAIN) {
        activeModule.activeSet.colorchain = tool.active;
      } else if (tool == TOOL_COLORLINK) {
        activeModule.activeSet.colorlink = tool.active;
      } else if (tool == TOOL_COLLAPSE) {
        activeModule.activeSet.collapse = tool.active;
      } else if (tool == TOOL_GLOBALMODE) {
        TOOL_COLORLINK.active  = activeModule.activeSet.colorlink;
        TOOL_COLORCHAIN.active = activeModule.activeSet.colorchain;
        TOOL_COLLAPSE.active   = activeModule.activeSet.collapse;
      } else if (tool == TOOL_CAPTURE) {
        if (logging && logActions) log.append(getTime() + " IMAGE CAPTURE CLICK\n");
        mode = CAPTURE;
        processCaptureClick();
      } else if (tool == TOOL_SAVE) {
        if (logging && logActions) log.append(getTime() + " SAVE ENTITY CLICK\n");
        mode = SAVE;
        processSaveClick();
      } else if (tool == TOOL_LOAD) {
        if (logging && logActions) log.append(getTime() + " LOAD ENTITY CLICK\n");
        mode = LOAD;
        processLoadClick();
      } else if (tool == TOOL_RANDOMIZE) {
        if (logging && logActions) log.append(getTime() + " RANDOMIZE CLICK\n");
        mode = PLAY;
        TOOL_RANDOMIZE.active = false;
        if (TOOL_GLOBALMODE.active) {
          for (EMCCVSet set : setList) {
            set.randomize();
          }
        } else {
          activeModule.activeSet.randomize();
        }
      } else if (tool == TOOL_RANDOMIZECOLOR) {
        if (logging && logActions) log.append(getTime() + " RANDOMIZE COLOR CLICK\n");
        mode = PLAY;
        TOOL_RANDOMIZECOLOR.active = false;
        processRandomizeColorClick();
        refreshModel = true;
      } else if (tool == TOOL_COLORSHIFT) {
        if (logging && logActions) log.append(getTime() + " COLOR SHIFT CLICK\n");
        mode = PLAY;
        TOOL_COLORSHIFT.active = false;
        processColorShiftClick();
        refreshModel = true;
        redrawAspectMenu = true;
      } else if (tool == TOOL_RELOAD) {
        if (logging && logActions) log.append(getTime() + " RELOAD CLICK\n");
        mode = RELOAD;
        processReloadClick();
      } else if (tool == TOOL_DASH) {
        if (logging && logActions) log.append(getTime() + " DASH CLICK\n");
        processDashClick();
      }

      redrawToolbar = true;
      processed = true;
      redraw();
    }
    if (processed) break;
  }
}

public void processRandomizeColorClick() {
  if (!TOOL_COLORLINK.active) {
    //run through every aspect in every set and apply a new random color
    if (TOOL_GLOBALMODE.active) {
      if (logging && logInfo) log.append(getTime() + "[RANDOMIZE COLORS] entered in GLOBAL MODE, no COLORLINK\n");
      for (EMCCVSet set : setList) {
        for (String aspectID : set.aspects) {
          EMCCVAspect aspect = (EMCCVAspect) elements.get(aspectID);
          if (aspect != null && aspect.randomizecolor) {
            aspect.colour = (color((int)random(255),(int)random(255),(int)random(255)));
          }
        }
      }
    } else {
      if (logging && logInfo) log.append(getTime() + "[RANDOMIZE COLORS] entered in TAB MODE, no COLORLINK\n");
      for (String aspectID : activeModule.activeSet.aspects) {
        EMCCVAspect aspect = (EMCCVAspect) elements.get(aspectID);
        if (aspect != null && aspect.randomizecolor) {
          aspect.colour = (color((int)random(255),(int)random(255),(int)random(255)));
        }
      }
    }
    
  } else {
    ArrayList oldcolors = new ArrayList(); // current colors
    ArrayList newcolors = new ArrayList(); // new random colors
    
    // find all colours currently in use and match it with a random one
    for (EMCCVFile file : fileLayers) {
      if (!oldcolors.contains(file.getColour())) {
        oldcolors.add(file.getColour()); // current color
        newcolors.add(color((int)random(255),(int)random(255),(int)random(255))); // corresponding random color
      }
    }
    
    //run through every aspect in every set and apply the matching random color
    if (TOOL_GLOBALMODE.active) {
      if (logging && logInfo) log.append(getTime() + "[RANDOMIZE COLORS] entered in GLOBAL MODE with COLORLINK\n");
      for (EMCCVSet set : setList) {
        for (String aspectID : set.aspects) {
          EMCCVAspect aspect = (EMCCVAspect) elements.get(aspectID);
          if (aspect != null && aspect.randomizecolor) {
            int i = oldcolors.indexOf(aspect.colour);
            aspect.colour = (int) newcolors.get(i);
          }
        }
      }
    } else {
      if (logging && logInfo) log.append(getTime() + "[RANDOMIZE COLORS] entered in TAB MODE with COLORLINK\n");
      for (String aspectID : activeModule.activeSet.aspects) {
        EMCCVAspect aspect = (EMCCVAspect) elements.get(aspectID);
        if (aspect != null && aspect.randomizecolor) {
          int i = oldcolors.indexOf(aspect.colour);
          aspect.colour = (int) newcolors.get(i);
        }
      }
    }
  }
}
public void processColorShiftClick() {
  ArrayList oldcolors = new ArrayList(); // current colors
  
  // find all colours currently in use and match it with a random one
  for (EMCCVFile file : fileLayers) {
    if (!oldcolors.contains(file.getColour())) {
      oldcolors.add(file.getColour()); // current color
    }
  }
  
  //run through every aspect in every set and shift the color
  if (TOOL_GLOBALMODE.active) {
    for (EMCCVSet set : setList) {
      for (String aspectID : set.aspects) {
        EMCCVAspect aspect = (EMCCVAspect) elements.get(aspectID);
        if (aspect != null && aspect.randomizecolor) {
          int i = oldcolors.indexOf(aspect.colour);
          aspect.colour = (int) oldcolors.get((i+1) % oldcolors.size());
        }
      }
    }
  } else {
    for (String aspectID : activeModule.activeSet.aspects) {
      EMCCVAspect aspect = (EMCCVAspect) elements.get(aspectID);
      if (aspect != null && aspect.randomizecolor) {
        int i = oldcolors.indexOf(aspect.colour);
        aspect.colour = (int) oldcolors.get((i+1) % oldcolors.size());
      }
    }
  }
}
public void processCaptureClick() {
  pgExport.save(sketchPath(DIR_IMAGES + getDateTime()  + ".png"));
}

public void processLoadClick() {
  if (TOOL_LOAD.active) {
    EMCCVThumb.thumbs.clear();

    File dir = new File(sketchPath(DIR_ENTITIES));

    File[] files = dir.listFiles();

    if (files != null) {
      if (logging && logInfo) log.append(getTime() + "[LOAD] Number of files found: " + files.length + "\n");
      for (File file : files) {
        String filename = file.getName();
        if (filename.contains(".png")) {
          EMCCVThumb thumb = new EMCCVThumb(filename.substring(0, filename.toString().lastIndexOf(".")));
          EMCCVThumb.thumbs.add(thumb);
        }
      }
    }
  } else {
    redrawAspectMenu = true;
    redrawModel = true;
  }
}

public void processReloadClick() {
  redrawSetMenu    = true;

  if (mode == DASH) {
    drawModuleMenu();
  } else {
    initXML(activeModule.id);
  }
  redraw();
}

public void processDashClick() {
  if (mode == DASH) {
    exit();
  } else {
    port_W = 0;
    port_H = 0;
    surface.setSize(800, 400);
    setList.clear();
    elements.clear();
    TOOL_LOAD.active = false;
    TOOL_DASH.active = true;
    redrawSetMenu    = true;
    redrawToolbar    = true;
    mode = DASH;
    redraw();
  }
}

public void processSaveClick() {
  String filename = getDateTime();
  PImage thumb = pg.copy();
  thumb.resize(0, 150);
  thumb.save(sketchPath(DIR_ENTITIES + filename  + ".png"));

  PrintWriter output = createWriter(sketchPath(DIR_ENTITIES + filename + ".txt"));

  for (EMCCVSet set : setList) {
    for (String aspectid : set.aspects) {
      StringBuilder sb = new StringBuilder();
      EMCCVAspect aspect = (EMCCVAspect) elements.get(aspectid);

      if (aspect.visible || aspect.alwayson) {
        sb.append(aspect.id).append(";").append(aspect.colour).append(";");

        for (Map.Entry<String, EMCCVStylegroup> groupentry : aspect.stylegroups.entrySet()) {
          for (Map.Entry<String, EMCCVStyle> styleentry : groupentry.getValue().styles.entrySet()) {
            if (styleentry.getValue().selected) {
              sb.append(styleentry.getValue().id).append(";");
              break;
            }
          }
        }
        output.println(sb.toString());
      }
    }
  }
  output.flush(); // Writes the remaining data to the file
  output.close(); // Finishes the file
}

public void processColorpickerClick() {

  ArrayList<EMCCVAspect> list = new ArrayList<EMCCVAspect>();

  boolean mouseOverColor = mouseOverHue() || mouseOverSat() || mouseOverBri() || mouseOverModel() || mouseOverTempColor() || mouseOverSwatch();

  // find all aspects with the same colour or part of the color chain
  if (mouseOverColor) {
    if (TOOL_COLORLINK.active || TOOL_COLORCHAIN.active) {

      if (TOOL_GLOBALMODE.active) {
        for (EMCCVSet set : setList) {
          for (String aspectid : set.aspects) {
            EMCCVAspect aspect = (EMCCVAspect) elements.get(aspectid);
            if (TOOL_COLORLINK.active && aspect.colour == activeAspect.colour) {
              list.add(aspect);
            } else if (TOOL_COLORCHAIN.active && activeAspect.colorchain && aspect.colorchain) {
              list.add(aspect);
            }
          }
        }
      } else {
        for (String aspectid : activeModule.activeSet.aspects) {
          EMCCVAspect aspect = (EMCCVAspect) elements.get(aspectid);
          if (TOOL_COLORLINK.active && aspect.colour == activeAspect.colour) {
            list.add(aspect);
          } else if (TOOL_COLORCHAIN.active && activeAspect.colorchain && aspect.colorchain) {
            list.add(aspect);
          }
        }
      }
    }
  }

  if (mouseOverHue() || mouseOverSat() || mouseOverBri()) {

    float hue = hue(activeAspect.colour);
    float sat = saturation(activeAspect.colour);
    float bri = brightness(activeAspect.colour);
    float value = mouseY - GRADIENT_Y;

    // HUE
    if (mouseOverHue()) {
      activeAspect.colour = color(value, sat, bri);
    }
    // SATURATION
    else if (mouseOverSat()) {
      activeAspect.colour = color(hue, value, bri);
    }
    // BRIGHTNESS
    else if (mouseOverBri()) {
      activeAspect.colour = color(hue, sat, value);
    }
  } else if (mouseOverModel() || mouseOverTempColor() || mouseOverSwatch()) {
    activeAspect.colour = get(mouseX, mouseY);
  }

  if ((TOOL_COLORLINK.active || TOOL_COLORCHAIN.active) && mouseOverColor) {
    for (EMCCVAspect aspect : list) {
      aspect.colour = activeAspect.colour;
    }
  }

  redrawModel = true;
  redraw();
}

public void processThumbnailClick() {
  for (EMCCVThumb thumb : EMCCVThumb.thumbs) {
    if (thumb.mouseOver(mouseX, mouseY)) {
      // A THUMBNAIL HAS BEEN CLICKED
      cursor(WAIT);
      if (logging && logActions) log.append(getTime() + "THUMB CLICK: " + thumb.SRC + "\n");

      // loading the txt file
      String[] stringlist = loadStrings(sketchPath(DIR_ENTITIES + thumb.SRC + ".txt"));

      if (stringlist.length > 0) {

        // turn all aspects off
        for (EMCCVSet set : setList) {
          for (String aspectid : set.aspects) {
            EMCCVAspect aspect = (EMCCVAspect) elements.get(aspectid);
            aspect.visible = false;
          }
        }

        // go through all lines in txt and apply them
        for (String s : stringlist) {
          String[] tokens = split(s, ";");
          EMCCVAspect aspect = (EMCCVAspect) elements.get(tokens[0]);
          if (aspect != null) {
            aspect.visible = true;
            try {
              if (tokens[1].charAt(0) == '#') {
                aspect.colour = unhex("FF" + tokens[1].substring(1));
              } else {
                aspect.colour = color(Integer.parseInt(tokens[1]));
              }
            } 
            catch (Exception e) {
              if (logging && logInfo) log.append(getTime() + "[LOADING ENTITY] ASPECT: " + aspect.id + "; COLOR INVALID\n");
            }
            for (int i = 2; i < tokens.length-1; i++) {
              EMCCVStyle style = (EMCCVStyle) elements.get(tokens[i]);
              if (style != null) {
                if (logging && logInfo) log.append(getTime() + "[LOADING ENTITY] ASPECT: " + aspect.id + "; STYLE: " + style.id + " (group: " + style.stylegroup + ")\n");
                EMCCVStylegroup group = aspect.stylegroups.get(style.stylegroup);
                group.resetStyles();
                style.selected = true;
              } else {
                log.append(getTime() + "[LOADING ENTITY] STYLE NOT FOUND: " + tokens[i] + "\n");
              }
            }
          } else {
            log.append(getTime() + "[LOADING ENTITY] ASPECT NOT FOUND: " + tokens[0] + "\n");
          }
        }
      }
    }
    // FINISHED LOADING ENTITY
    redrawAspectMenu = true;
    redrawModel      = true;
    refreshModel     = true;
    TOOL_LOAD.active = false;
    mode = PLAY;
    redraw();
  }
}


public boolean mouseOverHue() {
  if (mouseX >= GRADIENT_H_X && mouseX <= GRADIENT_H_X+GRADIENT_W &&
    mouseY >= GRADIENT_Y && mouseY <= GRADIENT_Y + GRADIENT_H) {
    return true;
  } else {
    return false;
  }
}

public boolean mouseOverSat() {
  if (mouseX >= GRADIENT_S_X && mouseX <= GRADIENT_S_X+GRADIENT_W &&
    mouseY >= GRADIENT_Y && mouseY <= GRADIENT_Y + GRADIENT_H) {
    return true;
  } else {
    return false;
  }
}

public boolean mouseOverBri() {
  if (mouseX >= GRADIENT_B_X && mouseX <= GRADIENT_B_X+GRADIENT_W &&
    mouseY >= GRADIENT_Y && mouseY <= GRADIENT_Y + GRADIENT_H) {
    return true;
  } else {
    return false;
  }
}

public boolean mouseOverModel() {
  if (mouseX >= IMG_X && mouseX <= IMG_X+view_W && 
    mouseY >= IMG_Y && mouseY <= IMG_Y+view_H) {
    return true;
  } else {
    return false;
  }
}

public boolean mouseOverTempColor() {
  if (mouseX >= GRADIENT_H_X && mouseX <= GRADIENT_H_X + 20 && 
    mouseY >= ASPECTMENU_Y + 19 && mouseY <= ASPECTMENU_Y + 39) {
    return true;
  } else {
    return false;
  }
}

public boolean mouseOverSwatch() {
  int columns = colorlist.size();
  if (columns > 10) {
    columns = 10;
  }
  if (mouseX >= SWATCH_X && mouseX <= SWATCH_X + (MARGIN+1)*(columns) && 
    mouseY >= SWATCH_Y && mouseY <= SWATCH_Y + (MARGIN+1)*(colorlist.size() / 10) + MARGIN) {
    return true;
  } else {
    return false;
  }
}
public static class EMCCVTool {
  
  // CONSTANTS
  static final int ICON_SIZE  = 32;
  static final int TOOLBAR_Y  = 60;
  static final int TOOLBAR_H  = ICON_SIZE + 4;
  static final int TOOLBAR_BG = 100;

  static final int ICON_ENABLED  = 128;
  static final int ICON_DISABLED = 50;
  static final int ICON_ACTIVE   = 255;
  static final int Y = (TOOLBAR_H - ICON_SIZE) / 2 + TOOLBAR_Y;

  // TOOL LIST
  static ArrayList<EMCCVTool> tools = new ArrayList<EMCCVTool>();

  // ATTRIBUTES
  final int X;
  final String title;
  final PImage ICON;
  boolean active;
  boolean disabled;

  // CONSTRUCTOR
  public EMCCVTool(final int X, String title, PImage ICON) {
    this.X = X;
    this.title = title;
    this.ICON = ICON;
    
    tools.add(this);
  }
  
  public boolean mouseOver(int x, int y) {
    if (x >= this.X && x <= this.X + ICON_SIZE && y >= Y && y <= Y + ICON_SIZE) {
      return true;
    } else {
      return false;
    }  
  }
  
  public int getColor() {
    if (this.disabled) {
      return ICON_DISABLED;
    } else if (this.active) {
      return ICON_ACTIVE;
    } else {
      return ICON_ENABLED;
    }
  }
}
  public void settings() {  size(800, 400); }
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "EMCCV_Prototype" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
