import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Random;
import java.util.Map.Entry;
import java.util.Map;
import java.util.LinkedHashMap;
import java.util.TreeMap;

String EMCCV_VERSION = "v4.2";
public static EMCCVModule activeModule = new EMCCVModule();
String CONFIGFILE;
String DIR_ENTITIES;
String DIR_MODULES = "modules/";
String DIR_EXT;
String DIR_MOD;
String DIR_IMAGES;
String DIR_DATA;

int mode = 0; // dashboard
final int DASH    = 0;
final int LOAD    = 1;
final int SAVE    = 2;
final int CAPTURE = 3;
final int RELOAD  = 4;
final int CP      = 5;
final int PLAY    = 6;

PGraphics pg;
PGraphics pgExport;

static Random random = new Random();

static XML xml;
static ArrayList<String> moduleids = new ArrayList();
static HashMap<String, EMCCVMenuElement> elements = new HashMap<String, EMCCVMenuElement>();
static Map<String, EMCCVMenuElement> randomKeys = new TreeMap<String, EMCCVMenuElement>();
static ArrayList<EMCCVFile> fileLayers = new ArrayList<EMCCVFile>();
static HashMap<String, PImage> graphics = new HashMap<String, PImage>();
static ArrayList COLOURS_50BRI = new ArrayList();

PFont OpenSans12;
PFont SansSerif12;

static boolean logging    = false;
static boolean logInitXML = logging && true;
static boolean logInfo    = logging && true;
static boolean logGui     = logging && true;
static boolean logActions = logging && false;
static boolean logSuccess = logging && false;
public static StringBuilder log = new StringBuilder();
static String logDirectory = "";
static String logline = "------------------------------------\n";
static String indent  = "||||||||||||||||||| ";

void setup() {
  size(800, 400);
  surface.setResizable(true);
  setLogDirectory();
  background(GUI_ACTIVE);
  
  //imageMode(CENTER);
  //textAlign(CENTER);
  //fill(GUI_TEXT);
  //text("loading...", width/2-100, height/2+50, 200, 50);
  //IMG_LOGO = loadImage(dataPath("images/gui/logo.png"));
  //image(IMG_LOGO, width/2, height/2);
  //imageMode(CORNER);
  //textAlign(CENTER, LEFT);
  
  colorMode(HSB, 255);
  ellipseMode(CORNER);
  noLoop();
  
  thread("loadIcons");
  thread("loadTools");
  thread("loadPalette");

  OpenSans12  = loadFont("OpenSans-12.vlw");
  SansSerif12 = loadFont("SansSerif-12.vlw");

  if (logging && logInfo) log.append(getTime() + "====================================\n");
  if (logging && logInfo) log.append(getTime() + " SETUP COMPLETE!\n");
  if (logging && logInfo) log.append(getTime() + "====================================\n");
  
  // Just in case we crash before the first draw finishes:
  log = printLog(log, "SETUP/");
}

void draw() {
  cursor(WAIT);
  noStroke();

  if (redrawSetMenu) drawMenuSets();
  if (redrawToolbar) drawToolbar();
  
  switch (mode) {
    case DASH:
      redrawAspectMenu = false;
      redrawModel      = false;
      drawModuleMenu();
      break;
      
    case CAPTURE:
      mode = PLAY;
      drawAlert("Image saved in the modules' images folder");
      TOOL_CAPTURE.active = false;
      redrawAspectMenu    = false;
      redrawToolbar       = true;
      frameRate(0.3);
      loop();
      redraw();
      break;
    case SAVE:
      mode = PLAY;
      drawAlert("Entity saved in the modules' entities folder");
      TOOL_SAVE.active = false;
      redrawAspectMenu = false;
      redrawToolbar    = true;
      frameRate(0.5);
      loop();
      redraw();
      break;
    case RELOAD:
      mode = PLAY;
      drawAlert("Reloading...");
      TOOL_RELOAD.active = false;
      redrawSetMenu      = true;
      redrawAspectMenu   = false;
      redrawToolbar      = true;
      frameRate(0.5);
      loop();
      redraw();
      break;
    case CP:
      drawColorPicker();
      break;
    case LOAD:
      redrawAspectMenu = false;
      redrawModel = false;
      loadThumbnails();
      break;
    default:
      if (refreshModel) {
        redrawAspectMenu = true;
        redrawModel      = true;
        refreshModel     = false;
        loop();
      } else {
        noLoop();
        frameRate(60);
        redrawAspectMenu = true;
        redrawModel = true;
      }
  }
  
  try {
    if (redrawAspectMenu && activeModule != null) drawAspectMenu();
    if (redrawModel) {
      updateView();
      fill(GUI_VIEW);
      rect(IMG_X, IMG_Y, width-ASPECTMENU_W, view_H);
      preparePG();
      resetCheckedMarker();
      //tint(0, 0, 0, 50);
      //image(pg, IMG_X+2*MARGIN, IMG_Y-MARGIN, view_W, view_H);
      noTint();
      image(pg, IMG_X, IMG_Y, view_W, view_H);
    }
    redrawModel = false;
    redrawSetMenu = false;
    redrawAspectMenu = false;
  } catch (Exception e) {
    log.append(getTime() + " ============================================================ \n");
    log.append(indent    + " ========= !!! ERROR DURING DRAWING PROCESS !!! ============= \n");
    log.append(indent    + " ========= !!!          Error #0001         !!! ============= \n");
    log.append(indent    + " ========= !!!       Reloading module       !!! ============= \n");
    log.append(indent    + " ============================================================ \n");
    redraw();
  }

  log = printLog(log, "DRAW/");
}

public void preparePG() {
  updateView();
  pg = createGraphics(activeModule.w, activeModule.h);
  pg.beginDraw();

  pg.colorMode(HSB, 255);
  pg.background(255, 255, 255, 0);
  pg.clear();
  pg.noStroke();
  pg.noTint();

  for (EMCCVFile file : fileLayers) {
    if (logging && logInfo) {
      log.append(getTime() + "[DRAW][LAYER]: ").append(file.layer).append(" - ").append(file.src).append("\n");  
      log.append(getTime() + "                 xmlParent: ").append(file.xmlParent).append("\n");
    }

    if (file.isActive()) {
      pg.blendMode(BLEND);
      
      // add image file to graphics HashMap if not there already
      if (graphics.get(sketchPath(DIR_DATA) + file.src) == null) {
        try {
          PImage img = loadImage(sketchPath(DIR_DATA) + file.src);
          graphics.put(sketchPath(DIR_DATA) + file.src, img);
          if (logging && logSuccess) log.append(getTime() + "[DRAW][FILE ADDED]: ").append(file.src).append("\n");
        } 
        catch (Exception e) {
          log.append(getTime() + "[DRAW] ==> FILE NOT FOUND: ").append(file.src).append("\n");
        }
      }

      // CHECKING FOR COLORMASTER
      if (file.colormaster != null && !file.colormaster.equals("")) {
        try {
          pg.tint(file.getColormaster().getColour());
        } 
        catch (Exception e) {
          log.append(getTime() + "[ERROR] Can't find colormaster." + e);
        }
      } else if (file.colorizable) {
        pg.tint(file.getColour());
      } else {
        pg.noTint();
      }
      if (file.blendmode != null && file.blendmode.equals("multiply")) {
        pg.blendMode(MULTIPLY);
      }
      try { //<>//
        // draw image
        if (file.w != 0 && file.h != 0) {
          // flip
          if (file.flipped) {
            pg.pushMatrix();
            pg.translate(graphics.get(sketchPath(DIR_DATA) + file.src).width,0);
            pg.scale(-1,1);
          }
          pg.image(graphics.get(sketchPath(DIR_DATA) + file.src), file.x, file.y, file.w, file.h);
          // unflip
          if (file.flipped) {
            pg.popMatrix();
          }
        } else {
          // flip
          if (file.flipped) {
            pg.pushMatrix();
            pg.translate(graphics.get(sketchPath(DIR_DATA) + file.src).width,0);
            pg.scale(-1,1);
          }
          pg.image(graphics.get(sketchPath(DIR_DATA) + file.src), file.x, file.y);
          // unflip
          if (file.flipped) {
            pg.popMatrix();
          }
        }
      } 
      catch (Exception e) {
        log.append("\n" + getTime() + "[ERROR] FILE COULD NOT BE FOUND: ").append(file.src).append("\n");
        println("\n" + getTime() + "[ERROR] FILE COULD NOT BE FOUND: " + file.src + "\n");
      }
    }
  }
  pg.endDraw();
  pgExport = createGraphics(activeModule.w, activeModule.h+20);
  pgExport.beginDraw();
  pgExport.image(pg, 0, 0);
  pgExport.noTint();
  pgExport.noStroke();

  color footerColor = pg.get(20, activeModule.h-20);
  pgExport.fill(footerColor);
  pgExport.rect(-1, activeModule.h-1, activeModule.w+2, 21);
  pgExport.fill(0, 0, 0, 75);
  pgExport.rect(-1, activeModule.h-1, activeModule.w+2, 22);
  pgExport.fill(255, 95);
  String credit = "Model: " + activeModule.title + " v" + activeModule.version;
  pgExport.textAlign(RIGHT, CENTER);
  pgExport.textFont(SansSerif12, 12);
  pgExport.text(credit, 0, activeModule.h, activeModule.w-5, 20);
  credit = "EMCCV Prototype " + EMCCV_VERSION;
  pgExport.textAlign(LEFT, CENTER);
  pgExport.text(credit, 5, activeModule.h, activeModule.w-5, 20);
  pgExport.noFill();

  pgExport.blendMode(BLEND);
  pgExport.image(IMG_CREDITS, 15, activeModule.h-50, 35, 35);
  pgExport.endDraw();
}

void resetCheckedMarker() {
  for (Map.Entry<String, EMCCVMenuElement> element : elements.entrySet()) {
    EMCCVMenuElement e = (EMCCVMenuElement) element.getValue();    
    e.checked = 0;
  }
}

void initXML(String moduleid) {
  
  randomKeys.clear();
  setList.clear();
  elements.clear();
  fileLayers.clear();
  graphics.clear();

  if (logging && logInitXML) {
    log.append(getTime() + "====================================\n");
    log.append(getTime() + " INITIALIZING XML\n");
    log.append(getTime() + "====================================\n");
  }
  
  activeModule    = new EMCCVModule();
  activeModule.id = moduleid;

  CONFIGFILE   = "modules/" + activeModule.id + "/config.xml";
  DIR_EXT      = "modules/" + activeModule.id + "/ext/";
  DIR_MOD      = "modules/" + activeModule.id + "/mod/";
  DIR_ENTITIES = "modules/" + activeModule.id + "/entities/";
  DIR_IMAGES   = "modules/" + activeModule.id + "/images/";
  DIR_DATA     = "modules/" + activeModule.id + "/data/";

  setLogDirectory();
  
  try {
    xml = loadXML(CONFIGFILE);
    activeModule.title    = xml.getString("title");
    activeModule.version  = xml.getString("version");
    activeModule.sDesc    = xml.getString("sDesc");
    activeModule.authorid = xml.getString("authorid");
    activeModule.w        = xml.getInt("width");
    activeModule.h        = xml.getInt("height");
    activeModule.logo     = loadImage(sketchPath(DIR_DATA + "logo.png"));
    activeModule.randomizecolor = xml.getString("randomizecolor") == null || xml.getString("randomizecolor").equals("true");
  
    if (logging && logInitXML) {
      log.append(getTime() + " WIDTH: ").append(activeModule.w).append(", HEIGHT: ").append(activeModule.h).append("\n");
      log.append(getTime() + logline);
    }
    
    EMCCVSet infoset = new EMCCVSet();
    infoset.id = "setInfo";
    infoset.title = "INFO";
    infoset.buttonwidth = 60;
    setList.add(infoset);
    if (logging && logInitXML) log.append(getTime() + " INFO tab added.\n");
  
    XML[] xmlSets = xml.getChildren("set");
  
    for (XML xmlSet : xmlSets) {
      EMCCVSet set = new EMCCVSet();
  
      // SET attributes
      set.id          = processStringAttribute(xmlSet, "id");
      set.title       = processStringAttribute(xmlSet, "title", "unknown title");
      set.disabled    = processBooleanAttribute(xmlSet, "disabled");
      set.parent      = processStringAttribute(xmlSet, "parent");
      set.colormaster = processStringAttribute(xmlSet, "colormaster");
      set.colour      = processColorAttribute(xmlSet, "color", color(255, 255, 255));
      set.colorizable = processBooleanAttribute(xmlSet, "colorizable");
      set.randomizecolor = processBooleanAttribute(xmlSet, "randomizecolor", activeModule.randomizecolor);
      set.x           = processIntAttribute(xmlSet, "x");
      set.y           = processIntAttribute(xmlSet, "y");
      set.w           = processIntAttribute(xmlSet, "w");
      set.h           = processIntAttribute(xmlSet, "h");
      set.buttonwidth = processIntAttribute(xmlSet, "buttonwidth");
      set.layer       = processFloatAttribute(xmlSet, "layer");
      set.alwayson    = processBooleanAttribute(xmlSet, "alwayson");
      set.randomize   = processStringAttribute(xmlSet, "randomize");
      set.probability = processIntAttribute(xmlSet, "probability", 75);
      set.blendmode   = processStringAttribute(xmlSet, "blendmode");
      set.flipped     = processBooleanAttribute(xmlSet, "flipped");
      set.visible     = processBooleanAttribute(xmlSet, "visible", false);
      set.incognito   = processBooleanAttribute(xmlSet, "incognito", false);
  
      if (logging && logInitXML) log.append(getTime() + logline);
  
      // SET > Aspects
      XML[] xmlAspects = xmlSet.getChildren("aspect");
      for (XML xmlAspect : xmlAspects) {
        EMCCVAspect aspect = new EMCCVAspect();
  
        // ASPECT attributes
        aspect.id          = processStringAttribute(xmlAspect, "id");      
        aspect.title       = processStringAttribute(xmlAspect, "title", "unknown title");
        aspect.disabled    = processBooleanAttribute(xmlAspect, "disabled");
        aspect.parent      = processStringAttribute(xmlAspect, "parent");
        aspect.colormaster = processStringAttribute(xmlAspect, "colormaster", set.colormaster);
        aspect.colour      = processColorAttribute(xmlAspect, "color", set.colour);
        aspect.colorizable = processBooleanAttribute(xmlAspect, "colorizable", set.colorizable);
        aspect.randomizecolor = processBooleanAttribute(xmlAspect, "randomizecolor", set.randomizecolor);
        aspect.x           = processIntAttribute(xmlAspect, "x", set.x);
        aspect.y           = processIntAttribute(xmlAspect, "y", set.y);
        aspect.w           = processIntAttribute(xmlAspect, "w", set.w);
        aspect.h           = processIntAttribute(xmlAspect, "h", set.h);
        aspect.visible     = processBooleanAttribute(xmlAspect, "visible", set.visible || processBooleanAttribute(xmlAspect, "alwayson", set.alwayson));
        aspect.incognito   = processBooleanAttribute(xmlAspect, "incognito", false);
        aspect.layer       = processFloatAttribute(xmlAspect, "layer", set.layer);
        aspect.alwayson    = processBooleanAttribute(xmlAspect, "alwayson", set.alwayson);
        aspect.randomize   = processStringAttribute(xmlAspect, "randomize");
        aspect.probability = processIntAttribute(xmlAspect, "probability", set.probability);
        aspect.blendmode   = processStringAttribute(xmlAspect, "blendmode", set.blendmode);
        aspect.flipped     = processBooleanAttribute(xmlAspect, "flipped", set.flipped);
        aspect.xmlParent   = set.id;
  
        if (aspect.alwayson) {
          aspect.visible = true;
        }
  
        if (logging && logInitXML) log.append(getTime() + logline);
  
        // SET > ASPECT > Styles
        XML[] xmlStyles = xmlAspect.getChildren("style");
        for (XML xmlStyle : xmlStyles) {
          EMCCVStyle style = new EMCCVStyle();
  
          // STYLE attributes
          style.id          = processStringAttribute(xmlStyle, "id");
          style.title       = processStringAttribute(xmlStyle, "title", "unknown title");
          style.disabled    = processBooleanAttribute(xmlStyle, "disabled");
          style.parent      = processStringAttribute(xmlStyle, "parent");
          style.colormaster = processStringAttribute(xmlStyle, "colormaster", aspect.colormaster);
          style.colour      = processColorAttribute(xmlStyle, "color", aspect.colour);
          style.colorizable = processBooleanAttribute(xmlStyle, "colorizable", aspect.colorizable);
          style.x           = processIntAttribute(xmlStyle, "x", aspect.x);
          style.y           = processIntAttribute(xmlStyle, "y", aspect.y);
          style.w           = processIntAttribute(xmlStyle, "w", aspect.w);
          style.h           = processIntAttribute(xmlStyle, "h", aspect.h);
          style.incognito   = processBooleanAttribute(xmlStyle, "incognito", false);
          style.layer       = processFloatAttribute(xmlStyle, "layer", aspect.layer);
          style.stylegroup  = processStringAttribute(xmlStyle, "stylegroup", "default");
          style.blendmode   = processStringAttribute(xmlStyle, "blendmode", aspect.blendmode);
          style.flipped     = processBooleanAttribute(xmlStyle, "flipped", aspect.flipped);
          style.selected    = processBooleanAttribute(xmlStyle, "selected", false);
          style.xmlParent   = aspect.id;
  
          // if the stylegroup doesn't exist -> create it
          if (aspect.stylegroups.get(style.stylegroup) == null) {
            EMCCVStylegroup stylegroup = new EMCCVStylegroup(style.stylegroup);
            aspect.stylegroups.put(stylegroup.title, stylegroup);
          }
  
          if (logging && logInitXML) log.append(getTime() + logline);
  
          // SET > ASPECT > STYLE > File
          XML[] xmlFiles = xmlStyle.getChildren("file");
          for (XML EMCCVFile : xmlFiles) {
            EMCCVFile file = new EMCCVFile();
  
            // FILE attributes
            file.parent      = processStringAttribute(EMCCVFile, "parent");
            file.colormaster = processStringAttribute(EMCCVFile, "colormaster", style.colormaster);
            file.colour      = processColorAttribute(EMCCVFile, "color", style.colour);
            file.colorizable = processBooleanAttribute(EMCCVFile, "colorizable", style.colorizable);
            file.x           = processIntAttribute(EMCCVFile, "x", style.x);
            file.y           = processIntAttribute(EMCCVFile, "y", style.y);
            file.w           = processIntAttribute(EMCCVFile, "w", style.w);
            file.h           = processIntAttribute(EMCCVFile, "h", style.h);
            file.layer       = processFloatAttribute(EMCCVFile, "layer", style.layer);
            file.src         = processStringAttribute(EMCCVFile, "src");
            file.blendmode   = processStringAttribute(EMCCVFile, "blendmode", style.blendmode);
            file.flipped     = processBooleanAttribute(EMCCVFile, "flipped", style.flipped);
            file.xmlParent   = style.id;
  
            // add file into ArrayList
            style.files.add(file);
            // add file into fileLayers ArrayList
            if (fileLayers.size() == 0 || file.layer >= fileLayers.get(fileLayers.size() - 1).layer) {
              fileLayers.add(file);
            } else {
              for (int i = 0; i < fileLayers.size(); i++) {
                if (fileLayers.get(i).layer > file.layer) {
                  fileLayers.add(i, file);
                  break;
                }
              }
            }
  
            if (logging && logInitXML) log.append(getTime() + logline);
          } // FILES DONE!
          if (logging && logInitXML) log.append(getTime() + " [INITXML] FILES added.\n");
          
          // add style into HashMaps and set the first one to selected
          if (aspect.stylegroups.get(style.stylegroup).styles.size() == 0) {
            style.selected = true;
          } else if (style.selected) {
            aspect.stylegroups.get(style.stylegroup).resetStyles();
          }
          aspect.stylegroups.get(style.stylegroup).styles.put(style.id, style);
          if (elements.get(style.id) != null) log.append(getTime() + " [INITXML] STYLE with ID = " + style.id + " already exists and will be overwritten!\n");
          elements.put(style.id, style);
        }
        if (logging && logInitXML) log.append(getTime() + " [INITXML] STYLES added.\n");
  
        // add aspect into HashMaps
        set.aspects.add(aspect.id);
        if (elements.get(aspect.id) != null) log.append(getTime() + " [INITXML] ASPECT with ID = " + aspect.id + " already exists and will be overwritten!\n");
        elements.put(aspect.id, aspect);
        if (aspect.randomize != null && !aspect.randomize.isEmpty()) {
          if (randomKeys.get(aspect.randomize) != null) log.append(getTime() + " [INITXML] ASPECT.randomize = " + aspect.randomize + " already exists and will be overwritten!\n");
          randomKeys.put(aspect.randomize, aspect);
        }
      }
      if (logging && logInitXML) log.append(getTime() + " [INITXML] ASPECTS added.\n");
      
      // add set into HashMap
      if (elements.get(set.id) != null) log.append(getTime() + " [INITXML] SET with ID = " + set.id + " already exists and will be overwritten!\n");
      elements.put(set.id, set);
      if (logging && logInitXML) log.append(getTime() + " [INITXML] SET " + set.id + " added to elements list.\n");
      
      if (set.randomize != null && !set.randomize.isEmpty()) {
        if (randomKeys.get(set.randomize) != null) log.append(getTime() + " [INITXML] SET.randomize = " + set.randomize + " already exists and will be overwritten!\n");
        randomKeys.put(set.randomize, set);
      }
      setList.add(set);
      if (logging && logInitXML) log.append(getTime() + " [INITXML] SET " + set.id + " added to setList.\n");
      
      activeModule.sets.add(set.id);
      if (logging && logInitXML) log.append(getTime() + " [INITXML] SET " + set.id + " added to activeModule.sets.\n");
    }
    if (logging && logInitXML) log.append(getTime() + " [INITXML] SETS added: " + setList.size()+ "\n");
    setList.get(0).selected = true;
    activeModule.activeSet = setList.get(0);
    if (logging && logInitXML) log.append(getTime() + " [INITXML] ACTIVE MODULE = " + setList.get(0).id + "\n");
  
    if (logging && logInitXML) {
      log.append(getTime() + "====================================\n");
      log.append(getTime() + " DATA ANALYSIS\n");
      log.append(getTime() + "====================================\n");
      log.append(getTime() + " Number of XML elements: ").append(elements.size()).append("\n");
      log.append(getTime() + " Number of fileLayers: ").append(fileLayers.size()).append("\n");
      log.append(getTime() + "\n");
    }
    log = printLog(log, "XML/");
  
    if (logging && logInitXML) {
      log.append(getTime() + "====================================\n");
      log.append(getTime() + " LIST OF LAYERS\n");
      log.append(getTime() + "====================================\n");
      log.append(getTime() + " Number of fileLayers: ").append(fileLayers.size()).append("\n");
      log.append(getTime() + "\n");
  
      for (EMCCVFile layer : fileLayers) {
        log.append(getTime() + "[LAYER " + layer.layer + "] " + layer.src + "\n");
      }
  
      log = printLog(log, "layers/");
    }
    
    updateView();
  
    int window_W = view_W + ASPECTMENU_W;
    int window_H = displayHeight - 120;
    surface.setSize(window_W, window_H);
  
  } catch (Exception e) {
    log.append(getTime() + "ERROR READING XML FILE\n");
  }
}

private String processStringAttribute(XML xml, String attribute) {
  String result = null;
  if (xml.getString(attribute) != null) {
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(xml.getString(attribute)).append("\n");
    result = xml.getString(attribute).trim().replaceAll(" +", " ");
  } else {
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" not declared.\n");
  }
  return result;
}

private String processStringAttribute(XML xml, String attribute, String inherited) {
  String result = null;
  if (xml.getString(attribute) != null) {
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(xml.getString(attribute)).append("\n");
    result = xml.getString(attribute).trim().replaceAll(" +", " ");
  } else if (inherited != null) {
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(inherited).append(" (inherited)\n");
    result = inherited;
  } else {
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" not declared.\n");
  }
  return result;
}

private boolean processBooleanAttribute(XML xml, String attribute) {
  boolean result = false;
  if (attribute == "colorizable") {
    result = true;
  }
  if (xml.getString(attribute) != null) {
    result = xml.getString(attribute).equals("true");
  }
  if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(result).append("\n");
  return result;
}

private boolean processBooleanAttribute(XML xml, String attribute, boolean inherited) {
  boolean result = false;
  if (attribute == "colorizable") {
    result = true;
  }
  if (xml.getString(attribute) != null) {
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(xml.getString(attribute)).append("\n");
    result = xml.getString(attribute).equals("true");
  } else {
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(inherited).append(" (inherited)\n");
    result = inherited;
  }
  return result;
}

private color processColorAttribute(XML xml, String attribute, color inherited) {
  color result = color(255);
  if (xml.getString(attribute) != null) {
    try {
      result = unhex("FF" + xml.getString(attribute).substring(1));
      if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(hex(result)).append("\n");
    } 
    catch (Exception e) {
      log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(hex(result)).append(" (declaration illegal)\n");
    }
  } else {
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(hex(inherited)).append(" (inherited)\n");
    result = inherited;
  }
  return result;
}

private float processFloatAttribute(XML xml, String attribute) {
  float result = 0;
  if (xml.getString(attribute) != null) {
    result = xml.getFloat(attribute);
  }
  if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(result).append("\n");
  return result;
}

private float processFloatAttribute(XML xml, String attribute, float inherited) {
  float result = 0;
  if (xml.getString(attribute) != null) {
    result = xml.getFloat(attribute);
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(xml.getString(attribute)).append("\n");
  } else {
    result = inherited;
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(inherited).append(" (inherited)\n");
  }
  return result;
}

private int processIntAttribute(XML xml, String attribute) {
  int result = 0;
  if (xml.getString(attribute) != null) {
    result = xml.getInt(attribute);
  }
  if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(result).append("\n");
  return result;
}

private int processIntAttribute(XML xml, String attribute, int inherited) {
  int result = 0;
  if (xml.getString(attribute) != null) {
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(xml.getString(attribute)).append("\n");
    result = xml.getInt(attribute);
  } else {
    if (logging && logInitXML) log.append(getTime() + xml.getName()).append(" ").append(attribute).append(" = ").append(inherited).append(" (inherited)\n");
    result = inherited;
  }
  return result;
}

/**
 * parses the parent attribute and checks to see if all conditions are met
 *
 */
public static boolean checkParents(String parents) {
  boolean bResult = true;

  if (parents != null) {
    // if the parent contains only one word, deal with it.
    if (!(parents.contains(",") || parents.contains(" "))) {
      EMCCVMenuElement parent;
      boolean invert = false;
      
      if (parents.contains("!")) {
        parent = elements.get(parents.substring(1));
        invert = true;
      } else {
        parent = elements.get(parents);
      }
      if (parent != null) {
        bResult = bResult && parent.isActive();
        if (invert) {
          bResult = !bResult;
        }
        if (logging && logInfo) log.append(getTime() + "[DRAW][IS ACTIVE][CHECK SINGLE PARENTS] " + parent.id + " active: " + bResult + "\n");
      } else {
        log.append(getTime() + parents).append(" id not a valid ID.\n");
      }
    } else {
      String[] prerequisites = split(parents, ", ");
      if (logging && logInfo) log.append(getTime() + "[DRAW][IS ACTIVE][CHECK MULTIPLE PARENTS] " + parents + ": " + prerequisites.length + " statement(s)\n");

      // iterate through prerequites
      for (String prereq : prerequisites) {

        // if this prerequisite contains multiple IDs (separated by spaces)...
        if (prereq.contains(" ")) {
          String[] words = split(prereq, " ");

          EMCCVMenuElement tempParent = null;
          boolean tempActive = true;
          String tempOperator = "";
          boolean prevWordWasId = false;

          if (logging && logInfo) log.append(getTime() + "      ");

          for (String word : words) {
            if (word.equals("and") || word.equals("or")) {
              if (prevWordWasId) {
                tempOperator = word;
                prevWordWasId = false;
                if (logging && logInfo) log.append(getTime() + " " + word + " ");
              } else {
                log.append(getTime() + "Expecting ID before operator. Prerequisite ignored.\n");
                break;
              }
            } else {
              // next word should be an ID -> deal with it.
              EMCCVMenuElement parent;
              boolean invert = false;
              
              if (word.contains("!")) {
                parent = elements.get(word.substring(1));
                invert = true;
              } else {
                parent = elements.get(word);
              }
              if (parent != null) {
                // we have a parent object
                if (logInfo && invert) log.append(getTime() + word + " (" + parent.isActive() + ")\n");
                if (logInfo && !invert) log.append(getTime() + word + " (" + !parent.isActive() + ")\n");

                if (tempParent == null) {
                  // this is the first ID in this statement -> memorize
                  tempParent = parent;
                  if (invert) {
                    tempActive = tempActive && !parent.isActive();
                  } else {
                    tempActive = tempActive && parent.isActive();
                  }                  
                  prevWordWasId = true;
                } else {
                  // we have already had an object in this statement
                  // -> we should also have an operator
                  if (!tempOperator.isEmpty()) {
                    // operator OK -> proceed
                    if (tempOperator.equals("and")) {
                      tempActive = tempActive && parent.isActive();
                    } else if (tempOperator.equals("or")) {
                      tempActive = tempActive || parent.isActive();
                    } else {
                      log.append(getTime() + "Expecting operator 'and/or' between IDs. Got ")
                        .append(word)
                        .append(" instead. Prerequisite ignored.\n");
                    }
                    tempOperator = "";
                    prevWordWasId = true;
                  } else {
                    log.append(getTime() + "Got multiple IDs (at ").append(word).append(") but no operator. Prerequisite ignored.\n");
                    break;
                  }
                }
              } else {
                log.append(getTime() + word).append(" id not a valid ID. Prerequisite ignored.\n");
                break;
              }
            }
          }
          bResult = bResult && tempActive;
        }
        if (logging && logInfo) log.append(getTime() + "[DRAW][IS ACTIVE][PREREQ] " + prereq + " (active: " + bResult + ")\n");
      }
    }
  }

  return bResult;
}

void drawAlert(String message) {
  textAlign(LEFT, TOP);
  fill(GUI_ACTIVE);
  rect(0, IMG_Y, ASPECTMENU_W, height-IMG_Y);
  fill(GUI_TEXT);
  text(message, MARGIN, IMG_Y+MARGIN, ASPECTMENU_W-2*MARGIN, height-IMG_Y-2*MARGIN);
}

static String getTime() {
  SimpleDateFormat sdf = new SimpleDateFormat("[yyyyMMdd-HH:mm:ss] ");
  return sdf.format(new Date());
}

static String getDateTime() {
  SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMddHHmmss");
  return sdf.format(new Date());
}

void loadModuleIDs() {
  moduleids.clear();
  
  File dir = new File(sketchPath(DIR_MODULES));
  File[] files = dir.listFiles();
  if (logging && logInfo) log.append(getTime() + "[LOAD MODULES] Number of files/directories found: " + files.length + "\n");

  moduleids = new ArrayList();

  if (files != null) {
    for (File file : files) {
      if (file.isDirectory()) {
        String filename = file.getName();
        if (logging && logInfo) log.append(getTime() + "[LOAD MODULES] Folder found: " + filename + "\n");
        File config = new File(sketchPath(DIR_MODULES + filename + "/config.xml"));
        if (config.isFile()) {
          moduleids.add(filename);
          if (logging && logInfo) log.append(getTime() + "[LOAD MODULES] Config found.\n");
        }
      }
    }
  }
}

void loadPalette() {
  COLOURS_50BRI.add(unhex("FFff0000"));
  COLOURS_50BRI.add(unhex("FFff4000"));
  COLOURS_50BRI.add(unhex("FFff8000"));
  COLOURS_50BRI.add(unhex("FFffbf00"));
  COLOURS_50BRI.add(unhex("FFffff00"));
  COLOURS_50BRI.add(unhex("FFbfff00"));
  COLOURS_50BRI.add(unhex("FF80ff00"));
  COLOURS_50BRI.add(unhex("FF40ff00"));
  COLOURS_50BRI.add(unhex("FF00ff00"));
  COLOURS_50BRI.add(unhex("FF00ff40"));
  COLOURS_50BRI.add(unhex("FF00ff80"));
  COLOURS_50BRI.add(unhex("FF00ffbf"));
  COLOURS_50BRI.add(unhex("FF00ffff"));
  COLOURS_50BRI.add(unhex("FF00bfff"));
  COLOURS_50BRI.add(unhex("FF0080ff"));
  COLOURS_50BRI.add(unhex("FF0040ff"));
  COLOURS_50BRI.add(unhex("FF0000ff"));
  COLOURS_50BRI.add(unhex("FF4000ff"));
  COLOURS_50BRI.add(unhex("FF8000ff"));
  COLOURS_50BRI.add(unhex("FFbf00ff"));
  COLOURS_50BRI.add(unhex("FFff00ff"));
  COLOURS_50BRI.add(unhex("FFff00bf"));
  COLOURS_50BRI.add(unhex("FFff0080"));
  COLOURS_50BRI.add(unhex("FFff0040"));
  COLOURS_50BRI.add(unhex("FFffffff"));
  COLOURS_50BRI.add(unhex("FFcccccc"));
  COLOURS_50BRI.add(unhex("FF999999"));
  COLOURS_50BRI.add(unhex("FF666666"));
  COLOURS_50BRI.add(unhex("FF333333"));
  COLOURS_50BRI.add(unhex("FF000000"));
  if (logging && logInfo) log.append(getTime() + "[SETUP] Palettes loaded.\n");
}
void loadIcons() {
  IMG_LOGO = loadImage(dataPath("images/gui/logo.png"));
  IMG_CREDITS =  loadImage(dataPath("images/gui/credits.png"));
  IMG_GRADIENT_HUE =  loadImage(dataPath("images/gui/gradient_hue.png"));
  IMG_GRADIENT_SAT =  loadImage(dataPath("images/gui/gradient_sat.png"));
  IMG_GRADIENT_BRI =  loadImage(dataPath("images/gui/gradient_bri.png"));
  ICON_CHECK    = loadImage(dataPath("images/icons/16/check.png"));
  ICON_LINK     = loadImage(dataPath("images/icons/16/link.png"));
  ICON_SETTINGS = loadImage(dataPath("images/icons/16/settings.png"));
  ICON_FIRST = loadImage(dataPath("images/icons/24/first.png"));
  ICON_UP    = loadImage(dataPath("images/icons/24/up.png"));
  ICON_DOWN  = loadImage(dataPath("images/icons/24/down.png"));
  ICON_LAST  = loadImage(dataPath("images/icons/24/last.png"));
  if (logging && logInfo) log.append(getTime() + "[SETUP] Icons loaded.\n");
}
void loadTools() {
  TOOL_LOAD           = new EMCCVTool(MARGIN + (EMCCVTool.ICON_SIZE + PADDING)*0, "Load Entity", loadImage(dataPath("images/icons/32/load.png")));
  TOOL_SAVE           = new EMCCVTool(MARGIN + (EMCCVTool.ICON_SIZE + PADDING)*1, "Save Entity", loadImage(dataPath("images/icons/32/save.png")));
  TOOL_CAPTURE        = new EMCCVTool(MARGIN + (EMCCVTool.ICON_SIZE + PADDING)*2, "Capture Image", loadImage(dataPath("images/icons/32/capture.png")));
  TOOL_COLORLINK      = new EMCCVTool(MARGIN*2 + (EMCCVTool.ICON_SIZE + PADDING)*3, "Color Link", loadImage(dataPath("images/icons/32/colorlink.png")));
  TOOL_COLORCHAIN     = new EMCCVTool(MARGIN*2 + (EMCCVTool.ICON_SIZE + PADDING)*4, "Color Chain", loadImage(dataPath("images/icons/32/colorchain.png")));
  TOOL_COLLAPSE       = new EMCCVTool(MARGIN*2 + (EMCCVTool.ICON_SIZE + PADDING)*5, "Collapse", loadImage(dataPath("images/icons/32/collapse.png")));
  TOOL_RANDOMIZE      = new EMCCVTool(MARGIN*2 + (EMCCVTool.ICON_SIZE + PADDING)*6, "Randomize", loadImage(dataPath("images/icons/32/randomize.png")));
  TOOL_RANDOMIZECOLOR = new EMCCVTool(MARGIN*2 + (EMCCVTool.ICON_SIZE + PADDING)*7, "Randomize Color", loadImage(dataPath("images/icons/32/randomizecolor.png")));
  TOOL_COLORSHIFT     = new EMCCVTool(MARGIN*2 + (EMCCVTool.ICON_SIZE + PADDING)*8, "Color Shoft", loadImage(dataPath("images/icons/32/colorshift.png")));
  TOOL_GLOBALMODE     = new EMCCVTool(MARGIN*2 + (EMCCVTool.ICON_SIZE + PADDING)*9, "Global Mode", loadImage(dataPath("images/icons/32/globalmode.png")));
  TOOL_LOG            = new EMCCVTool(MARGIN*3 + (EMCCVTool.ICON_SIZE + PADDING)*10, "Log", loadImage(dataPath("images/icons/32/log.png")));
  TOOL_RELOAD         = new EMCCVTool(MARGIN*3 + (EMCCVTool.ICON_SIZE + PADDING)*11, "Reload", loadImage(dataPath("images/icons/32/reload.png")));
  TOOL_DASH           = new EMCCVTool(MARGIN*3 + (EMCCVTool.ICON_SIZE + PADDING)*12, "Dash", loadImage(dataPath("images/icons/32/close.png")));
  
  TOOL_LOG.active = false;
  if (logging && logInfo) log.append(getTime() + "[SETUP] Tools loaded.\n");
}
void loadThumbnails() {
  fill(GUI_ACTIVE);
  rect(0, IMG_Y, width, height-IMG_Y);

  PImage img;
  int x = MARGIN;
  int y = IMG_Y + MARGIN;

  if (logging && logActions) log.append(getTime() + "[DRAW][LOAD] Number of thumbs in list: " + EMCCVThumb.thumbs.size() + "\n");

  for (int i = 0; i < EMCCVThumb.thumbs.size(); i++) {

    EMCCVThumb thumb = EMCCVThumb.thumbs.get(i);
    thumb.x = x;
    thumb.y = y;
    img = loadImage(sketchPath(DIR_ENTITIES + thumb.SRC + ".png"));

    if (logging && logGui) log.append(getTime() + "[DRAW][LOAD] drawing: SRC = " + thumb.SRC + "\n");
    if (logging && logGui) log.append(getTime() + "[DRAW][LOAD]            x = " + thumb.x + "\n");
    if (logging && logGui) log.append(getTime() + "[DRAW][LOAD]            y = " + thumb.y + "\n");

    fill(GUI_VIEW);
    rect(x, y, EMCCVThumb.W, EMCCVThumb.H);
    noTint();
    image(img, x, y);
    //fill(GUI_TEXT);
    //textAlign(CENTER);
    //text(thumb.SRC, x, y, EMCCVThumb.W, TEXTSIZE*2);

    x = x + EMCCVThumb.W + 2;

    if (x > width-EMCCVThumb.W-MARGIN) {
      if (logging && logGui) log.append(getTime() + "[DRAW][LOAD] NEW ROW\n");
      x = MARGIN;
      y += (EMCCVThumb.H + 2);
    }
  }
}

public StringBuilder printLog(StringBuilder log, String directoy) {
  if (log.length() > 0) {
        
    String filename = "logs/"+ logDirectory + "/"  + directoy + getDateTime() + ".txt";
    
    PrintWriter output = createWriter(filename);
    output.println(log);
    output.flush(); // Write the remaining data
    output.close(); // Finish the file
    //println(log);
  }
  return new StringBuilder();
}

public static void setLogDirectory() {
  SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMdd");
  
  if (activeModule.id != null && !activeModule.id.isEmpty()) {
    logDirectory = activeModule.id + "/" + sdf.format(new Date());
  } else {
    logDirectory = "_dashboard/" + sdf.format(new Date());
  }
}