// GUI CONSTANTS
final static int PADDING      = 10;
final static int MARGIN       = 20;
final static int MENURADIUS   = 4;
final static int TEXTSIZE     = 13;
final static int CONTROLSIZE  = 16;

// TOOLS
public static EMCCVTool TOOL_LOAD;
public static EMCCVTool TOOL_SAVE;
public static EMCCVTool TOOL_CAPTURE;
public static EMCCVTool TOOL_COLORLINK;
public static EMCCVTool TOOL_COLORCHAIN;
public static EMCCVTool TOOL_COLLAPSE;
public static EMCCVTool TOOL_GLOBALMODE;
public static EMCCVTool TOOL_RANDOMIZE;
public static EMCCVTool TOOL_LOG;
public static EMCCVTool TOOL_RELOAD;
public static EMCCVTool TOOL_DASH;
public static EMCCVTool TOOL_COLORSHIFT;
public static EMCCVTool TOOL_RANDOMIZECOLOR;

// IMAGES
public static PImage IMG_LOGO;
public static PImage IMG_CREDITS;
public static PImage IMG_GRADIENT_HUE;
public static PImage IMG_GRADIENT_SAT;
public static PImage IMG_GRADIENT_BRI;
public static PImage ICON_CHECK;
public static PImage ICON_LINK;
public static PImage ICON_SETTINGS;

public static PImage ICON_FIRST;
public static PImage ICON_UP;
public static PImage ICON_DOWN;
public static PImage ICON_LAST;

// SETS
final static int MENUSET_W = 70;
final static int MENUSET_H = 19;
final static int MENUSET_Y = EMCCVTool.TOOLBAR_Y + EMCCVTool.TOOLBAR_H + MARGIN;
final static int MENUSET_D = 2;
final static int MENUSET_P = 10;

// ASPECTS
final static int ASPECTMENU_W = 285;
final static int ASPECTMENU_X = MARGIN;
final static int ASPECTMENU_Y = MENUSET_Y + MENUSET_H + ASPECTMENU_X;
final static int ASPECTMENU_D = 22;

// COLOR PICKER
final static int GRADIENT_W   = 65;
final static int GRADIENT_H   = 255;
final static int GRADIENT_H_X = ASPECTMENU_X;
final static int GRADIENT_S_X = ASPECTMENU_X + GRADIENT_W + PADDING;
final static int GRADIENT_B_X = ASPECTMENU_X + GRADIENT_W*2 + PADDING*2;
final static int GRADIENT_Y   = ASPECTMENU_Y + GRADIENT_W + MARGIN;

final static int SWATCH_X     = GRADIENT_H_X;
final static int SWATCH_Y     = GRADIENT_Y + GRADIENT_H + 80;
static ArrayList colorlist = new ArrayList();

final static int CP_X         = ASPECTMENU_X+19;
final static int LINK_X       = ASPECTMENU_X+38;
final static int COG_X        = ASPECTMENU_X+57;

public static int getAspectTitleX() {
  if (TOOL_COLORCHAIN.active) {
    return COG_X+19;
  }
  return LINK_X+19;
}
public static int getAspectTitleW() {
  return ASPECTMENU_W - getAspectTitleX();
}

final static int IMG_X = ASPECTMENU_W;
final static int IMG_Y = MENUSET_H + MENUSET_Y;

int currentWidth = width;
int currentHeight = height;
int port_W = displayWidth;
int port_H = displayHeight;
float factor = 1.0;
int view_W = displayWidth;
int view_H = displayHeight;


// COLORS
final static color GUI_ENABLED      = 0xFF646464; // color(100, 100, 100);
final static color GUI_DISABLED     = 0x50646464; // color(100, 100, 100, 80);
final static color GUI_ACTIVE       = 0xFF808080; // color(128, 128, 128);
final static color GUI_VIEW         = 0xFF878787; // color(135, 135, 135);
final static color GUI_SELECTED     = 0xFFB4B4B4; // color(180, 180, 180);
final static color GUI_SELECTED_DIS = 0x50B4B4B4; // color(180, 180, 180, 80);
final static color GUI_BG           = 0xFF505050; // color(80, 80, 80);
final static color GUI_TEXT         = 0xFFDCDCDC; // color(220, 220, 220);
final static color GUI_TEXT_DIS     = 0x50DCDCDC; // color(220, 220, 220, 80);

// GRADIENT
int Y_AXIS = 1;
int X_AXIS = 2;

// GUI STATES
static ArrayList<EMCCVSet> setList = new ArrayList<EMCCVSet>();
static ArrayList<EMCCVButton> buttonList = new ArrayList<EMCCVButton>();
static EMCCVAspect activeAspect;

static boolean redrawModel = true;
static boolean refreshModel = false;
static boolean redrawSetMenu = true;
static boolean redrawToolbar = true;
static boolean redrawAspectMenu = true;
static color tempColor = 0;
static boolean bTempColor = false;


//////////////////////////////////////////////////////////////////////////////////////////////// UPDATE VIEW

public void updateView() {
  if (port_W != 0) {
    port_W = width - IMG_X;
  } else port_W = displayWidth - IMG_X - 20;

  if (port_H != 0) {
    port_H = height - IMG_Y;
  } else port_H = displayHeight - IMG_Y - 120;

  view_W = floor(activeModule.w * factor);
  view_H = floor(activeModule.h * factor);

  if (logging && logGui) {
    log.append(getTime() + "====================================\n");
    log.append(getTime() + "[UPDATEVIEW BEFORE] width: "  + width  + "\n");
    log.append(getTime() + "[UPDATEVIEW BEFORE] height: " + height + "\n");
    log.append(getTime() + "[UPDATEVIEW BEFORE] port_W: " + port_W + "\n");
    log.append(getTime() + "[UPDATEVIEW BEFORE] port_H: " + port_H + "\n");
    log.append(getTime() + "[UPDATEVIEW BEFORE] view_W: " + view_W + "\n");
    log.append(getTime() + "[UPDATEVIEW BEFORE] view_H: " + view_H + "\n");
    log.append(getTime() + "[UPDATEVIEW BEFORE] factor: " + factor + "\n");
    log.append(getTime() + "[UPDATEVIEW BEFORE] factor_W: " + (float) port_W / (float) activeModule.w + "\n");
    log.append(getTime() + "[UPDATEVIEW BEFORE] factor_H: " + (float) port_H / (float) activeModule.h + "\n");
  }
  float factor_W = (float) port_W / (float) activeModule.w;
  float factor_H = (float) port_H / (float) activeModule.h;

  if (factor_W < factor_H) {
    factor = factor_W;
  } else {
    factor = factor_H;
  }

  if (factor > 1) {
    factor = 1;
  }

  view_W = floor(activeModule.w * factor);
  view_H = floor(activeModule.h * factor);

  if (logging && logGui) {
    log.append(getTime() + "[UPDATEVIEW AFTER] view_W: " + view_W + "\n");
    log.append(getTime() + "[UPDATEVIEW AFTER] view_H: " + view_H + "\n");
    log.append(getTime() + "[UPDATEVIEW AFTER] factor_W: " + factor_W + "\n");
    log.append(getTime() + "[UPDATEVIEW AFTER] factor_H: " + factor_H + "\n");
    log.append(getTime() + "[UPDATEVIEW AFTER] factor: " + factor + "\n");
    log.append(getTime() + "====================================\n");
  }
}


//////////////////////////////////////////////////////////////////////////////////////////////// DRAW SET MENU

void drawMenuSets() {
  fill(GUI_BG);
  rect(0, 0, width, MENUSET_H + MENUSET_Y);
  textAlign(CENTER, CENTER);

  int x = MENUSET_D;
  for (EMCCVSet set : setList) {

    if (!set.incognito) {
      if (set.selected) {
        fill(GUI_ACTIVE);
        activeModule.activeSet = set;
      } else {
        fill(GUI_ENABLED);
      }

      rect(x, MENUSET_Y, set.buttonwidth, MENUSET_H, MENURADIUS, MENURADIUS, 0, 0);
      fill(GUI_TEXT);
      text(set.title, x, MENUSET_Y-2, set.buttonwidth, MENUSET_H);

      EMCCVButton button = new EMCCVButton(x, MENUSET_Y, set.buttonwidth, MENUSET_H, set);
      buttonList.add(button);

      x += set.buttonwidth + MENUSET_D;
    }
  }
  noTint();
  image(IMG_LOGO, width-PADDING-254, (EMCCVTool.TOOLBAR_Y-40)/2);

  textAlign(LEFT, CENTER);
  fill(GUI_TEXT);
  textSize(16);
  text(activeModule.title, PADDING, 0, width-PADDING-254-PADDING*2, EMCCVTool.TOOLBAR_Y);
  textSize(TEXTSIZE);
}

//////////////////////////////////////////////////////////////////////////////////////////////// DRAW TOOLBAR
public void drawToolbar() {  
  fill(255, 0, EMCCVTool.TOOLBAR_BG);
  rect(0, EMCCVTool.TOOLBAR_Y, width, EMCCVTool.TOOLBAR_H );

  for (EMCCVTool tool : EMCCVTool.tools) {
    tint(255, tool.getColor());
    image(tool.ICON, tool.X, EMCCVTool.Y);
  }
}


//////////////////////////////////////////////////////////////////////////////////////////////// DRAW ASPECT MENU

public void drawAspectMenu() {

  resetButtonList();

  // ASPECT MENU SETUP
  fill(GUI_ACTIVE);
  rect(0, IMG_Y, ASPECTMENU_W, height - IMG_Y); // background
  textAlign(LEFT, CENTER);
  //int o = activeModule.activeSet.offset; // offset for scrolling  

  // DRAWING THE MENU
  if(setList.isEmpty()) {
    log.append(getTime() + " [DRAWING ASPECT MENU] ERROR: The active MODULE seems to have no sets.\n");
  } else if (activeModule.activeSet == null) {
    activeModule.activeSet = (EMCCVSet) elements.get(setList.get(0).title);
    log.append(getTime() + " [DRAWING ASPECT MENU] ERROR: The active MODULE was null. Something is wrong here...\n");
  }
    if (setList.get(0) == activeModule.activeSet) {
      
      drawInfoTab();
      
    } else {
      activeModule.activeSet.drawAspectMenu(this);
    }
    noTint();
}


//////////////////////////////////////////////////////////////////////////////////////////////// DRAW MODULE MENU

void drawModuleMenu() {
  fill(GUI_ACTIVE);
  rect(0, IMG_Y, width, height-IMG_Y);

  int w = 100;
  int h = 40;
  int x = MARGIN;
  int y = IMG_Y+MARGIN;
  int xcount = floor((width-MARGIN*2) / (w+PADDING));

  loadModuleIDs();
  if (moduleids.size() > 0) {

    for (int i = 0; i < moduleids.size(); i++) {
      x = MARGIN       + ((i * (w + PADDING)) % (xcount * (w+PADDING)));
      if ((i+1) % (xcount+1) == 0) {
        y = y + h + PADDING;
      }
      //y = IMG_Y+MARGIN + ((i * (h + PADDING)) / (ycount * (h+PADDING)));

      fill(GUI_ENABLED);
      rect(x, y, w, h, 3);
      fill(GUI_TEXT);
      text(moduleids.get(i), x+PADDING, y, w-2*PADDING, h);

      // MODULE into elements list
      EMCCVModule module = new EMCCVModule();
      module.id = moduleids.get(i);
      elements.put(module.id, module);

      // BUTTON into buttonlist
      EMCCVButton modulebutton = new EMCCVButton(x, y, w, h, "module", module);
      buttonList.add(modulebutton);
    }
  } else {
    fill(GUI_ENABLED);
    rect(x, y, w, h, 3);
    fill(GUI_TEXT);
    text("No modules found.", x+PADDING, y, w-2*PADDING, h);
  }
}